# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

import numpy
from daCore import BasicObjects, NumericObjects
from daAlgorithms.Atoms import ecwblue

# ==============================================================================
class ElementaryAlgorithm(BasicObjects.Algorithm):
    def __init__(self):
        BasicObjects.Algorithm.__init__(self, "BLUE")
        self.defineRequiredParameter(
            name     = "Variant",
            default  = "Blue",
            typecast = str,
            message  = "Variant ou formulation de la méthode",
            listval  = [
                "Blue",
                ],
            listadv  = [
                "OneCorrection",
                ],
            )
        self.defineRequiredParameter(
            name     = "EstimationOf",
            default  = "Parameters",
            typecast = str,
            message  = "Estimation d'état ou de paramètres",
            listval  = ["State", "Parameters"],
            )
        self.defineRequiredParameter(
            name     = "StoreInternalVariables",
            default  = False,
            typecast = bool,
            message  = "Stockage des variables internes ou intermédiaires du calcul",
            )
        self.defineRequiredParameter(
            name     = "StoreSupplementaryCalculations",
            default  = [],
            typecast = tuple,
            message  = "Liste de calculs supplémentaires à stocker et/ou effectuer",
            listval  = [
                "Analysis",
                "APosterioriCorrelations",
                "APosterioriCovariance",
                "APosterioriStandardDeviations",
                "APosterioriVariances",
                "BMA",
                "CostFunctionJ",
                "CostFunctionJAtCurrentOptimum",
                "CostFunctionJb",
                "CostFunctionJbAtCurrentOptimum",
                "CostFunctionJo",
                "CostFunctionJoAtCurrentOptimum",
                "CurrentOptimum",
                "CurrentState",
                "CurrentStepNumber",
                "ForecastState",
                "Innovation",
                "InnovationAtCurrentAnalysis",
                "MahalanobisConsistency",
                "OMA",
                "OMB",
                "SampledStateForQuantiles",
                "SigmaBck2",
                "SigmaObs2",
                "SimulatedObservationAtBackground",
                "SimulatedObservationAtCurrentOptimum",
                "SimulatedObservationAtCurrentState",
                "SimulatedObservationAtOptimum",
                "SimulationQuantiles",
                ]
            )
        self.defineRequiredParameter(
            name     = "Quantiles",
            default  = [],
            typecast = tuple,
            message  = "Liste des valeurs de quantiles",
            minval   = 0.,
            maxval   = 1.,
            )
        self.defineRequiredParameter(
            name     = "SetSeed",
            typecast = numpy.random.seed,
            message  = "Graine fixée pour le générateur aléatoire",
            )
        self.defineRequiredParameter(
            name     = "NumberOfSamplesForQuantiles",
            default  = 100,
            typecast = int,
            message  = "Nombre d'échantillons simulés pour le calcul des quantiles",
            minval   = 1,
            )
        self.defineRequiredParameter(
            name     = "SimulationForQuantiles",
            default  = "Linear",
            typecast = str,
            message  = "Type de simulation en estimation des quantiles",
            listval  = ["Linear", "NonLinear"]
            )
        self.defineRequiredParameter( # Pas de type
            name     = "StateBoundsForQuantiles",
            message  = "Liste des paires de bornes pour les états utilisés en estimation des quantiles",
            )
        self.requireInputArguments(
            mandatory= ("Xb", "Y", "HO", "R", "B"),
            optional = ("U", "EM", "CM", "Q"),
            )
        self.setAttributes(tags=(
            "DataAssimilation",
            "Linear",
            "Filter",
            ))

    def run(self, Xb=None, Y=None, U=None, HO=None, EM=None, CM=None, R=None, B=None, Q=None, Parameters=None):
        self._pre_run(Parameters, Xb, Y, U, HO, EM, CM, R, B, Q)
        #
        #--------------------------
        if   self._parameters["Variant"] == "Blue":
            NumericObjects.multiXOsteps(self, Xb, Y, U, HO, EM, CM, R, B, Q, ecwblue.ecwblue)
        #
        #--------------------------
        elif self._parameters["Variant"] == "OneCorrection":
            ecwblue.ecwblue(self, Xb, Y, U, HO, CM, R, B)
        #
        #--------------------------
        else:
            raise ValueError("Error in Variant name: %s"%self._parameters["Variant"])
        #
        self._post_run(HO)
        return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This file is part of SALOME ADAO module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

import argparse # optparse deprecated since Python version 3.2
import sys
import re

import Traducteur.log as log
from Traducteur.load         import getJDC, getJDCFromTexte
from Traducteur.mocles       import parseKeywords
from Traducteur.dictErreurs  import genereErreurPourCommande
from Traducteur.inseremocle  import *
from Traducteur.movemocle    import *
from Traducteur.renamemocle  import *

version_out = "V9_9_0"

usage="""Usage: python %(prog)s [args]

Typical use is:
  python %(prog)s --infile=xxxx.comm --outfile=yyyy.comm"""

atraiter = (
    "ASSIMILATION_STUDY",
    "CHECKING_STUDY",
    )

dict_erreurs = {
    "ASSIMILATION_STUDY":"Changements dans l'arbre et dans les noms",
    "CHECKING_STUDY":"Changements dans l'arbre et dans les noms",
    }

sys.dict_erreurs=dict_erreurs

def traduc(infile=None,outfile=None,texte=None,flog=None):
    hdlr = log.initialise(flog)
    if infile is not None:
        jdc  = getJDC(infile,atraiter)
    elif texte is not None:
        jdc  = getJDCFromTexte(texte,atraiter)
    else:
        raise ValueError("Traduction du JDC impossible")

    #Parse les mocles des commandes
    parseKeywords(jdc.root)
    genereErreurPourCommande(jdc,('Algorithm','AlgorithmParameters','FunctionDict'))
    # ==========================================================================

    for command in atraiter:
        # Insere le MC s'il n'existe pas
        chercheOperInsereFacteurSiRegle(jdc, command, "AlgorithmParameters",((("AlgorithmParameters",),"nexistepasMCFParmi"),))
        # Deplace le MC
        moveMotClefInOperToFact(jdc, command, "Algorithm", "AlgorithmParameters", plusieursFois=False)
        # Renomme le MC
        renameMotCleInFact(jdc, command, "AlgorithmParameters", "INPUT_TYPE", "Parameters")
        # Renomme le MC
        renameMotCle(jdc, command, "Study_name", "StudyName")
        renameMotCle(jdc, command, "Study_repertory", "StudyRepertory")

    # ==========================================================================
    fsrc = jdc.getSource()
    fsrc = re.sub( "MaximumNumberOfSteps", "MaximumNumberOfIterations", fsrc )
    fsrc = re.sub( "FunctionDict", "ScriptWithSwitch", fsrc )
    fsrc = re.sub( "FUNCTIONDICT_FILE", "SCRIPTWITHSWITCH_FILE", fsrc )
    fsrc = re.sub( "#VERSION_CATALOGUE:.*:FIN VERSION_CATALOGUE", "#VERSION_CATALOGUE:%s:FIN VERSION_CATALOGUE"%version_out, fsrc)
    fsrc = re.sub( "#CHECKSUM.*FIN CHECKSUM", "", fsrc )
    #
    log.ferme(hdlr)
    if outfile is not None:
        with open(outfile,'w') as f:
            f.write( fsrc )
    else:
        return fsrc

class MonTraducteur:
    def __init__(self,texte):
        self.__texte = str(texte)
    def traduit(self):
        return traduc(infile=None,outfile=None,texte=self.__texte,flog=None)

def main():
    parser = argparse.ArgumentParser(usage=usage)

    parser.add_argument('-i','--infile', dest="infile",
        help="Le fichier COMM en entree, a traduire")
    parser.add_argument('-o','--outfile', dest="outfile", default='out.comm',
        help="Le fichier COMM en sortie, traduit")

    args = parser.parse_args()
    if len(args.infile) == 0:
        print("")
        parser.print_help()
        print("")
        sys.exit(1)

    traduc(args.infile,args.outfile)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# Copyright (C) 2007-2008  CEA/DEN, EDF R&D, OPEN CASCADE
#
# Copyright (C) 2003-2007  OPEN CASCADE, EADS/CCR, LIP6, CEA/DEN,
# CEDRAT, EDF R&D, LEG, PRINCIPIA R&D, BUREAU VERITAS
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#

__author__ = "gboulant"
__date__ = "$1 avr. 2010 09:08:02$"

class Enumerate(object):
    """
    This class emulates a C-like enum for python.
    """
    def __init__(self, keys, offset=0):
        """
        Canonical constructor
        @keys a list of keys string to be used as the enum symbols
        """
        self._dict_keynumbers = {}
        for number, key in enumerate(keys):
            value = offset+number
            setattr(self, key, value)
            self._dict_keynumbers[key] = value


    def contains(self, key):
        """
        Return true if this enumerate contains the specified key
        @key a key to test
        """
        return (key in self._dict_keynumbers.keys())

    def isValid(self, value):
        return (value in self._dict_keynumbers.values())

    def listkeys(self):
        return sorted(self._dict_keynumbers.keys())

    def listvalues(self):
        return sorted(self._dict_keynumbers.values())

#
# ==============================================================================
# Basic use cases and unit test functions
# ==============================================================================
#

def TEST_simple():
    TYPES_LIST=Enumerate([
        'SEP',
        'OTHER'
    ])
    print(TYPES_LIST.listvalues())
    return True

def TEST_createFromList():
    codes = Enumerate([
        'KERNEL', # This should take the value 0
        'GUI', # This should take the value 1
        'GEOM', # ...
        'MED',
        'SMESH'])

    print(codes.KERNEL)
    print(codes.GEOM)
    if (codes.KERNEL == 0 and codes.GEOM == 2):
        return True
    else:
        return False

def TEST_createFromString():
    aList = "KERNEL GUI GEOM MED"

    codes = Enumerate(aList.split())

    print(codes.KERNEL)
    print(codes.GEOM)
    if (codes.KERNEL == 0 and codes.GEOM == 2):
        return True
    else:
        return False

def TEST_contains():
    codes = Enumerate([
        'KERNEL', # This should take the value 0
        'GUI', # This should take the value 1
        'GEOM', # ...
        'MED',
        'SMESH'])

    print("VISU in enumerate?", codes.contains("VISU"))
    if ( not codes.contains("VISU") ):
        return True
    else:
        return False

def TEST_isValid():
    codes = Enumerate([
        'KERNEL', # This should take the value 0
        'GUI', # This should take the value 1
        'GEOM', # ...
        'MED',
        'SMESH'])

    if ( not codes.isValid(23) ):
        return True
    else:
        return False

def TEST_offset():
    codes = Enumerate([
        'KERNEL', # This should take the value 0
        'GUI', # This should take the value 1
        'GEOM', # ...
        'MED',
        'SMESH'], offset=20)

    print(codes.KERNEL)
    print(codes.GEOM)
    if (codes.KERNEL == 20 and codes.GEOM == 22):
        return True
    else:
        return False

def TEST_listvalues():
    codes = Enumerate([
        'KERNEL', # This should take the value 0
        'GUI', # This should take the value 1
        'GEOM', # ...
        'MED',
        'SMESH'], offset=20)

    print(codes.listvalues())
    if codes.listvalues() != [20,21,22,23,24]:
        return False
    return True


if __name__ == "__main__":
    import unittester
    unittester.run("enumerate","TEST_simple")
    unittester.run("enumerate","TEST_createFromList")
    unittester.run("enumerate","TEST_createFromString")
    unittester.run("enumerate","TEST_contains")
    unittester.run("enumerate","TEST_isValid")
    unittester.run("enumerate","TEST_offset")
    unittester.run("enumerate","TEST_listvalues")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This file is part of SALOME ADAO module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: André Ribes, andre.ribes@edf.fr, EDF R&D

from generator.generator_python import PythonGenerator
import logging

def entryPoint():
   """
      Retourne les informations necessaires pour le chargeur de plugins

      Ces informations sont retournees dans un dictionnaire
   """
   return {
        # Le nom du plugin
        'name' : 'adao',
        # La factory pour creer une instance du plugin
          'factory' : AdaoGenerator,
          }

class AdaoGenerator(PythonGenerator):

  def __init__(self,cr=None):
    PythonGenerator.__init__(self, cr)
    self.dictMCVal={}
    self.text_comm = ""
    self.text_da = ""
    self.text_da_status = False
    self.logger = logging.getLogger('ADAO EFICAS GENERATOR')
    self.logger.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    formatter = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
    ch.setFormatter(formatter)
    self.logger.addHandler(ch)

  def gener(self,obj,format='brut',config=None,appliEficas=None):
    self.logger.debug("method gener called")
    self.text_comm = PythonGenerator.gener(self, obj, format, config, appliEficas)
    for key, value in self.dictMCVal.items():
      self.logger.debug("dictMCVAl %s %s" % (key,value))

    try :
      self.text_da_status = False
      self.generate_da()
      self.text_da_status = True
    except:
      self.logger.debug("EFICAS case is not valid, python command file for YACS schema generation cannot be created")
      self.logger.debug(self.text_da)
      self.dictMCVal = {}
      # import traceback ; traceback.print_exc()
    return self.text_comm

  def writeDefault(self, fn):
    if self.text_da_status:
      self.logger.debug("write adao python command file")
      filename = fn[:fn.rfind(".")] + '.py'
      #~ f = open( str(filename), 'wb')
      f = open( str(filename), 'w')
      f.write( self.text_da )
      f.close()

  def generMCSIMP(self,obj) :
    """
    Convertit un objet MCSIMP en texte python
    """
    clef=""
    for i in obj.getGenealogie() :
      clef=clef+"__"+i
    self.dictMCVal[clef]=obj.valeur

    s=PythonGenerator.generMCSIMP(self,obj)
    return s

  def generate_da(self):

    self.text_da += "#-*- coding: utf-8 -*-\n"
    self.text_da += "study_config = {}\n"

    # Extraction de Study_type
    if "__CHECKING_STUDY__StudyName" in self.dictMCVal.keys():
      self.type_of_study = "CHECKING_STUDY"
      self.text_da += "study_config['StudyType'] = '\"CHECKING_STUDY\"'\n"
    elif "__OPTIMIZATION_STUDY__StudyName" in self.dictMCVal.keys():
      self.type_of_study = "OPTIMIZATION_STUDY"
      self.text_da += "study_config['StudyType'] = '\"ASSIMILATION_STUDY\"'\n"
    elif "__REDUCTION_STUDY__StudyName" in self.dictMCVal.keys():
      self.type_of_study = "REDUCTION_STUDY"
      self.text_da += "study_config['StudyType'] = '\"ASSIMILATION_STUDY\"'\n"
    else:
      self.type_of_study = "ASSIMILATION_STUDY"
      self.text_da += "study_config['StudyType'] = '\"ASSIMILATION_STUDY\"'\n"

    # Extraction de StudyName
    self.text_da += "study_config['Name'] = '" + self.dictMCVal["__"+self.type_of_study+"__StudyName"] + "'\n"
    # Extraction de Debug
    if "__"+self.type_of_study+"__Debug" in self.dictMCVal.keys():
      self.text_da += "study_config['Debug'] = '" + str(self.dictMCVal["__"+self.type_of_study+"__Debug"]) + "'\n"
    else:
      self.text_da += "study_config['Debug'] = '0'\n"

    # Extraction de Algorithm et de ses parametres
    if "__"+self.type_of_study+"__AlgorithmParameters__Algorithm" in self.dictMCVal.keys():
      self.text_da += "study_config['Algorithm'] = '" + self.dictMCVal["__"+self.type_of_study+"__AlgorithmParameters__Algorithm"] + "'\n"
      self.add_AlgorithmParameters()
    elif "__"+self.type_of_study+"__Algorithm" in self.dictMCVal.keys():
      self.text_da += "study_config['Algorithm'] = '" + self.dictMCVal["__"+self.type_of_study+"__Algorithm"] + "'\n"

    if "__"+self.type_of_study+"__Background__INPUT_TYPE" in self.dictMCVal.keys():
      self.add_data("Background")
    if "__"+self.type_of_study+"__BackgroundError__INPUT_TYPE" in self.dictMCVal.keys():
      self.add_data("BackgroundError")
    if "__"+self.type_of_study+"__Observation__INPUT_TYPE" in self.dictMCVal.keys():
      self.add_data("Observation")
    if "__"+self.type_of_study+"__ObservationError__INPUT_TYPE" in self.dictMCVal.keys():
      self.add_data("ObservationError")
    if "__"+self.type_of_study+"__CheckingPoint__INPUT_TYPE" in self.dictMCVal.keys():
      self.add_data("CheckingPoint")
    if "__"+self.type_of_study+"__ObservationOperator__INPUT_TYPE" in self.dictMCVal.keys():
      self.add_data("ObservationOperator")
    if "__"+self.type_of_study+"__EvolutionModel__INPUT_TYPE" in self.dictMCVal.keys():
      self.add_data("EvolutionModel")
    if "__"+self.type_of_study+"__EvolutionError__INPUT_TYPE" in self.dictMCVal.keys():
      self.add_data("EvolutionError")
    if "__"+self.type_of_study+"__ControlInput__INPUT_TYPE" in self.dictMCVal.keys():
      self.add_data("ControlInput")

    self.add_variables()
    # Parametres optionnels

    # Extraction du StudyRepertory
    if "__"+self.type_of_study+"__StudyRepertory" in self.dictMCVal.keys():
      self.text_da += "study_config['Repertory'] = '" + self.dictMCVal["__"+self.type_of_study+"__StudyRepertory"] + "'\n"

    # Extraction du ExecuteInContainer
    if "__"+self.type_of_study+"__ExecuteInContainer" in self.dictMCVal.keys():
      self.text_da += "study_config['ExecuteInContainer'] = '" + str(self.dictMCVal["__"+self.type_of_study+"__ExecuteInContainer"]) + "'\n"
    else:
      self.text_da += "study_config['ExecuteInContainer'] = 'No'\n"

    # Extraction de UserPostAnalysis
    if "__"+self.type_of_study+"__UserPostAnalysis__FROM" in self.dictMCVal.keys():
      self.add_UserPostAnalysis()
    if "__"+self.type_of_study+"__UserDataInit__INIT_FILE" in self.dictMCVal.keys():
      self.add_init()
    if "__"+self.type_of_study+"__Observers__SELECTION" in self.dictMCVal.keys():
      self.add_observers()

  def add_data(self, data_name):

    # Extraction des donnees
    search_text = "__"+self.type_of_study+"__" + data_name + "__"
    data_type = self.dictMCVal[search_text + "INPUT_TYPE"]
    search_type = search_text + data_type + "__data__"
    from_type = self.dictMCVal[search_type + "FROM"]
    data = ""
    if from_type == "String":
      data = self.dictMCVal[search_type + "STRING_DATA__STRING"]
    elif from_type == "Script":
      data = self.dictMCVal[search_type + "SCRIPT_DATA__SCRIPT_FILE"]
    elif from_type == "DataFile":
      data = self.dictMCVal[search_type + "DATA_DATA__DATA_FILE"]
    elif from_type == "ScriptWithSwitch":
      data = self.dictMCVal[search_type + "SCRIPTWITHSWITCH_DATA__SCRIPTWITHSWITCH_FILE"]
    elif from_type == "ScriptWithFunctions":
      data = self.dictMCVal[search_type + "SCRIPTWITHFUNCTIONS_DATA__SCRIPTWITHFUNCTIONS_FILE"]
    elif from_type == "ScriptWithOneFunction":
      data = self.dictMCVal[search_type + "SCRIPTWITHONEFUNCTION_DATA__SCRIPTWITHONEFUNCTION_FILE"]
    elif from_type == "FunctionDict":
      data = self.dictMCVal[search_type + "FUNCTIONDICT_DATA__FUNCTIONDICT_FILE"]
    else:
      raise Exception('From Type unknown', from_type)

    if from_type == "String" or from_type == "Script" or from_type == "DataFile":
      self.text_da += data_name + "_config = {}\n"
      self.text_da += data_name + "_config['Type'] = '" + data_type + "'\n"
      self.text_da += data_name + "_config['From'] = '" + from_type + "'\n"
      self.text_da += data_name + "_config['Data'] = '" + data      + "'\n"
      if search_text+"Stored" in self.dictMCVal.keys():
        self.text_da += data_name + "_config['Stored'] = '" +  str(self.dictMCVal[search_text+"Stored"])  + "'\n"
      if search_type+"DATA_DATA__ColMajor" in self.dictMCVal.keys():
        self.text_da += data_name + "_config['ColMajor'] = '" +  str(self.dictMCVal[search_type+"DATA_DATA__ColMajor"])  + "'\n"
      self.text_da += "study_config['" + data_name + "'] = " + data_name + "_config\n"

    if from_type == "ScriptWithSwitch":
      self.text_da += data_name + "_ScriptWithSwitch = {}\n"
      self.text_da += data_name + "_ScriptWithSwitch['Function'] = ['Direct', 'Tangent', 'Adjoint']\n"
      self.text_da += data_name + "_ScriptWithSwitch['Script'] = {}\n"
      self.text_da += data_name + "_ScriptWithSwitch['Script']['Direct'] = '"  + data + "'\n"
      self.text_da += data_name + "_ScriptWithSwitch['Script']['Tangent'] = '" + data + "'\n"
      self.text_da += data_name + "_ScriptWithSwitch['Script']['Adjoint'] = '" + data + "'\n"
      self.text_da += data_name + "_config = {}\n"
      self.text_da += data_name + "_config['Type'] = 'Function'\n"
      self.text_da += data_name + "_config['From'] = 'ScriptWithSwitch'\n"
      self.text_da += data_name + "_config['Data'] = " + data_name + "_ScriptWithSwitch\n"
      self.text_da += "study_config['" + data_name + "'] = " + data_name + "_config\n"

    if from_type == "ScriptWithFunctions":
      self.text_da += data_name + "_ScriptWithFunctions = {}\n"
      self.text_da += data_name + "_ScriptWithFunctions['Function'] = ['Direct', 'Tangent', 'Adjoint']\n"
      self.text_da += data_name + "_ScriptWithFunctions['Script'] = {}\n"
      self.text_da += data_name + "_ScriptWithFunctions['Script']['Direct'] = '"  + data + "'\n"
      self.text_da += data_name + "_ScriptWithFunctions['Script']['Tangent'] = '" + data + "'\n"
      self.text_da += data_name + "_ScriptWithFunctions['Script']['Adjoint'] = '" + data + "'\n"
      self.text_da += data_name + "_config = {}\n"
      self.text_da += data_name + "_config['Type'] = 'Function'\n"
      self.text_da += data_name + "_config['From'] = 'ScriptWithFunctions'\n"
      self.text_da += data_name + "_config['Data'] = " + data_name + "_ScriptWithFunctions\n"
      self.text_da += "study_config['" + data_name + "'] = " + data_name + "_config\n"

    if from_type == "ScriptWithOneFunction":
      self.text_da += data_name + "_ScriptWithOneFunction = {}\n"
      self.text_da += data_name + "_ScriptWithOneFunction['Function'] = ['Direct', 'Tangent', 'Adjoint']\n"
      self.text_da += data_name + "_ScriptWithOneFunction['Script'] = {}\n"
      self.text_da += data_name + "_ScriptWithOneFunction['Script']['Direct'] = '"  + data + "'\n"
      self.text_da += data_name + "_ScriptWithOneFunction['Script']['Tangent'] = '" + data + "'\n"
      self.text_da += data_name + "_ScriptWithOneFunction['Script']['Adjoint'] = '" + data + "'\n"
      self.text_da += data_name + "_ScriptWithOneFunction['DifferentialIncrement'] = " + str(float(self.dictMCVal[search_type + "SCRIPTWITHONEFUNCTION_DATA__DifferentialIncrement"])) + "\n"
      self.text_da += data_name + "_ScriptWithOneFunction['CenteredFiniteDifference'] = " + str(self.dictMCVal[search_type + "SCRIPTWITHONEFUNCTION_DATA__CenteredFiniteDifference"]) + "\n"
      if search_type + "SCRIPTWITHONEFUNCTION_DATA__EnableMultiProcessing" in self.dictMCVal.keys():
        self.text_da += data_name + "_ScriptWithOneFunction['EnableMultiProcessing'] = " + str(self.dictMCVal[search_type + "SCRIPTWITHONEFUNCTION_DATA__EnableMultiProcessing"]) + "\n"
      if search_type + "SCRIPTWITHONEFUNCTION_DATA__NumberOfProcesses" in self.dictMCVal.keys():
        self.text_da += data_name + "_ScriptWithOneFunction['NumberOfProcesses'] = " + str(self.dictMCVal[search_type + "SCRIPTWITHONEFUNCTION_DATA__NumberOfProcesses"]) + "\n"
      self.text_da += data_name + "_config = {}\n"
      self.text_da += data_name + "_config['Type'] = 'Function'\n"
      self.text_da += data_name + "_config['From'] = 'ScriptWithOneFunction'\n"
      self.text_da += data_name + "_config['Data'] = " + data_name + "_ScriptWithOneFunction\n"
      self.text_da += "study_config['" + data_name + "'] = " + data_name + "_config\n"

    if from_type == "FunctionDict":
      self.text_da += data_name + "_FunctionDict = {}\n"
      self.text_da += data_name + "_FunctionDict['Function'] = ['Direct', 'Tangent', 'Adjoint']\n"
      self.text_da += data_name + "_FunctionDict['Script'] = {}\n"
      self.text_da += data_name + "_FunctionDict['Script']['Direct'] = '"  + data + "'\n"
      self.text_da += data_name + "_FunctionDict['Script']['Tangent'] = '" + data + "'\n"
      self.text_da += data_name + "_FunctionDict['Script']['Adjoint'] = '" + data + "'\n"
      self.text_da += data_name + "_config = {}\n"
      self.text_da += data_name + "_config['Type'] = 'Function'\n"
      self.text_da += data_name + "_config['From'] = 'FunctionDict'\n"
      self.text_da += data_name + "_config['Data'] = " + data_name + "_FunctionDict\n"
      self.text_da += "study_config['" + data_name + "'] = " + data_name + "_config\n"

  def add_init(self):

      init_file_data = self.dictMCVal["__"+self.type_of_study+"__UserDataInit__INIT_FILE"]
      init_target_list = self.dictMCVal["__"+self.type_of_study+"__UserDataInit__TARGET_LIST"]

      self.text_da += "Init_config = {}\n"
      self.text_da += "Init_config['Type'] = 'Dict'\n"
      self.text_da += "Init_config['From'] = 'Script'\n"
      self.text_da += "Init_config['Data'] = '" + init_file_data + "'\n"
      self.text_da += "Init_config['Target'] = ["
      if isinstance(init_target_list, "str"):
        self.text_da +=  "'" + init_target_list + "',"
      else:
        for target in init_target_list:
          self.text_da += "'" + target + "',"
      self.text_da += "]\n"
      self.text_da += "study_config['UserDataInit'] = Init_config\n"

  def add_UserPostAnalysis(self):

    from_type = self.dictMCVal["__"+self.type_of_study+"__UserPostAnalysis__FROM"]
    data = ""
    if from_type == "String":
      data = self.dictMCVal["__"+self.type_of_study+"__UserPostAnalysis__STRING_DATA__STRING"]
      self.text_da += "Analysis_config = {}\n"
      self.text_da += "Analysis_config['From'] = 'String'\n"
      self.text_da += "Analysis_config['Data'] = \"\"\"" + data + "\"\"\"\n"
      self.text_da += "study_config['UserPostAnalysis'] = Analysis_config\n"
    elif from_type == "Script":
      data = self.dictMCVal["__"+self.type_of_study+"__UserPostAnalysis__SCRIPT_DATA__SCRIPT_FILE"]
      self.text_da += "Analysis_config = {}\n"
      self.text_da += "Analysis_config['From'] = 'Script'\n"
      self.text_da += "Analysis_config['Data'] = '" + data + "'\n"
      self.text_da += "study_config['UserPostAnalysis'] = Analysis_config\n"
    elif from_type == "Template":
      tmpl = self.dictMCVal["__"+self.type_of_study+"__UserPostAnalysis__TEMPLATE_DATA__Template"]
      data = self.dictMCVal["__"+self.type_of_study+"__UserPostAnalysis__TEMPLATE_DATA__%s__ValueTemplate"%tmpl]
      self.text_da += "Analysis_config = {}\n"
      self.text_da += "Analysis_config['From'] = 'String'\n"
      self.text_da += "Analysis_config['Data'] = \"\"\"" + data + "\"\"\"\n"
      self.text_da += "study_config['UserPostAnalysis'] = Analysis_config\n"
    else:
      raise Exception('From Type unknown', from_type)

  def add_AlgorithmParameters(self):

    data_name = "AlgorithmParameters"
    data_type = "Dict"
    #
    if "__"+self.type_of_study+"__AlgorithmParameters__Parameters" in self.dictMCVal:
        para_type = self.dictMCVal["__"+self.type_of_study+"__AlgorithmParameters__Parameters"]
    else:
        para_type = "Defaults"
    #
    if para_type == "Defaults":
        from_type = para_type
    elif para_type == "Dict":
        from_type = self.dictMCVal["__"+self.type_of_study+"__AlgorithmParameters__Dict__data__FROM"]
    else:
        return

    if from_type == "Script":
      data = self.dictMCVal["__"+self.type_of_study+"__AlgorithmParameters__Dict__data__SCRIPT_DATA__SCRIPT_FILE"]
      self.text_da += data_name + "_config = {}\n"
      self.text_da += data_name + "_config['Type'] = '" + data_type + "'\n"
      self.text_da += data_name + "_config['From'] = '" + from_type + "'\n"
      self.text_da += data_name + "_config['Data'] = '" + data + "'\n"
      self.text_da += "study_config['" + data_name + "'] = " + data_name + "_config\n"
    elif from_type == "String":
      data = self.dictMCVal["__"+self.type_of_study+"__AlgorithmParameters__Dict__data__STRING_DATA__STRING"]
      self.text_da += data_name + "_config = {}\n"
      self.text_da += data_name + "_config['Type'] = '" + data_type + "'\n"
      self.text_da += data_name + "_config['From'] = '" + from_type + "'\n"
      self.text_da += data_name + "_config['Data'] = '" + data + "'\n"
      self.text_da += "study_config['" + data_name + "'] = " + data_name + "_config\n"
    elif from_type == "Defaults":
      base = "__"+self.type_of_study+"__AlgorithmParameters__Parameters"
      keys = [k for k in self.dictMCVal.keys() if base in k]
      if base in keys: keys.remove(base)
      keys = [k.replace(base,'') for k in keys]
      data  = '{'
      for k in keys:
        key = k.split('__')[-1]
        val = self.dictMCVal[base+k]
        # print key," = ",val,"    ",type(val)
        if isinstance(val, str) and key == "SetSeed":
            data += '"%s":%s,'%(key,int(val))
        elif isinstance(val, str) and not (val.count('[')>=2 or val.count('(')>=2):
            data += '"%s":"%s",'%(key,val)
        else:
            data += '"%s":%s,'%(key,val)
      data = data.replace("'",'"')
      data += '}'
      if data != '{}':
          self.text_da += data_name + "_config = {}\n"
          self.text_da += data_name + "_config['Type'] = '" + data_type + "'\n"
          self.text_da += data_name + "_config['From'] = 'String'\n"
          self.text_da += data_name + "_config['Data'] = '" + data + "'\n"
          self.text_da += "study_config['" + data_name + "'] = " + data_name + "_config\n"

  def add_variables(self):

    # Input variables
    if "__"+self.type_of_study+"__InputVariables__NAMES" in self.dictMCVal.keys():
      names = []
      sizes = []
      if isinstance(self.dictMCVal["__"+self.type_of_study+"__InputVariables__NAMES"], type("")):
        names.append(self.dictMCVal["__"+self.type_of_study+"__InputVariables__NAMES"])
      else:
        names = self.dictMCVal["__"+self.type_of_study+"__InputVariables__NAMES"]
      if isinstance(self.dictMCVal["__"+self.type_of_study+"__InputVariables__SIZES"], type(1)):
        sizes.append(self.dictMCVal["__"+self.type_of_study+"__InputVariables__SIZES"])
      else:
        sizes = self.dictMCVal["__"+self.type_of_study+"__InputVariables__SIZES"]

      self.text_da += "inputvariables_config = {}\n"
      self.text_da += "inputvariables_config['Order'] = %s\n" % list(names)
      for name, size in zip(names, sizes):
        self.text_da += "inputvariables_config['%s'] = %s\n" % (name,size)
      self.text_da += "study_config['InputVariables'] = inputvariables_config\n"
    else:
      self.text_da += "inputvariables_config = {}\n"
      self.text_da += "inputvariables_config['Order'] =['adao_default']\n"
      self.text_da += "inputvariables_config['adao_default'] = -1\n"
      self.text_da += "study_config['InputVariables'] = inputvariables_config\n"

    # Output variables
    if "__"+self.type_of_study+"__OutputVariables__NAMES" in self.dictMCVal.keys():
      names = []
      sizes = []
      if isinstance(self.dictMCVal["__"+self.type_of_study+"__OutputVariables__NAMES"], type("")):
        names.append(self.dictMCVal["__"+self.type_of_study+"__OutputVariables__NAMES"])
      else:
        names = self.dictMCVal["__"+self.type_of_study+"__OutputVariables__NAMES"]
      if isinstance(self.dictMCVal["__"+self.type_of_study+"__OutputVariables__SIZES"], type(1)):
        sizes.append(self.dictMCVal["__"+self.type_of_study+"__OutputVariables__SIZES"])
      else:
        sizes = self.dictMCVal["__"+self.type_of_study+"__OutputVariables__SIZES"]

      self.text_da += "outputvariables_config = {}\n"
      self.text_da += "outputvariables_config['Order'] = %s\n" % list(names)
      for name, size in zip(names, sizes):
        self.text_da += "outputvariables_config['%s'] = %s\n" % (name,size)
      self.text_da += "study_config['OutputVariables'] = outputvariables_config\n"
    else:
      self.text_da += "outputvariables_config = {}\n"
      self.text_da += "outputvariables_config['Order'] = ['adao_default']\n"
      self.text_da += "outputvariables_config['adao_default'] = -1\n"
      self.text_da += "study_config['OutputVariables'] = outputvariables_config\n"

  def add_observers(self):
    observers = {}
    observer = self.dictMCVal["__"+self.type_of_study+"__Observers__SELECTION"]
    if isinstance(observer, type("")):
      self.add_observer_in_dict(observer, observers)
    else:
      for observer in self.dictMCVal["__"+self.type_of_study+"__Observers__SELECTION"]:
        self.add_observer_in_dict(observer, observers)

    # Write observers in the python command file
    number = 2
    self.text_da += "observers = {}\n"
    for observer in observers.keys():
      number += 1
      self.text_da += "observers[\"" + observer + "\"] = {}\n"
      self.text_da += "observers[\"" + observer + "\"][\"number\"] = " + str(number) + "\n"
      self.text_da += "observers[\"" + observer + "\"][\"nodetype\"] = \"" + observers[observer]["nodetype"] + "\"\n"
      if observers[observer]["nodetype"] == "String":
        self.text_da += "observers[\"" + observer + "\"][\"String\"] = \"\"\"" + observers[observer]["script"] + "\"\"\"\n"
      elif observers[observer]["nodetype"] == "Template":
        self.text_da += "observers[\"" + observer + "\"][\"String\"] = \"\"\"" + observers[observer]["script"] + "\"\"\"\n"
        self.text_da += "observers[\"" + observer + "\"][\"Template\"] = \"\"\"" + observers[observer]["template"] + "\"\"\"\n"
      else:
        self.text_da += "observers[\"" + observer + "\"][\"Script\"] = \"" + observers[observer]["file"] + "\"\n"
      if "scheduler" in observers[observer].keys():
        self.text_da += "observers[\"" + observer + "\"][\"scheduler\"] = \"\"\"" + observers[observer]["scheduler"] + "\"\"\"\n"
      if "info" in observers[observer].keys():
        self.text_da += "observers[\"" + observer + "\"][\"info\"] = \"\"\"" + observers[observer]["info"] + "\"\"\"\n"
    self.text_da += "study_config['Observers'] = observers\n"

  def add_observer_in_dict(self, observer, observers):
    """
      Add observer in the observers dict.
    """
    observers[observer] = {}
    observers[observer]["name"] = observer
    observer_eficas_name = "__"+self.type_of_study+"__Observers__" + observer + "__" + observer + "_data__"
    # NodeType
    node_type_key_name = observer_eficas_name + "NodeType"
    observers[observer]["nodetype"] = self.dictMCVal[node_type_key_name]

    # NodeType script/file
    if observers[observer]["nodetype"] == "String":
      observers[observer]["script"] = self.dictMCVal[observer_eficas_name + "PythonScript__Value"]
    elif observers[observer]["nodetype"] == "Template":
      observers[observer]["nodetype"] = "String"
      observer_template_key = observer_eficas_name + "ObserverTemplate__"
      observers[observer]["template"] = self.dictMCVal[observer_template_key + "Template"]
      observers[observer]["script"]   = self.dictMCVal[observer_template_key + observers[observer]["template"] + "__ValueTemplate"]
    else:
      observers[observer]["file"] = self.dictMCVal[observer_eficas_name + "UserFile__Value"]

    # Scheduler
    scheduler_key_name = observer_eficas_name + "Scheduler"
    if scheduler_key_name in self.dictMCVal.keys():
      observers[observer]["scheduler"] = self.dictMCVal[scheduler_key_name]

    # Info
    info_key_name = observer_eficas_name + "Info"
    if info_key_name in self.dictMCVal.keys():
      observers[observer]["info"] = self.dictMCVal[info_key_name]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

__doc__ = """
    Constrained (2UKF) Unscented Kalman Filter
"""
__author__ = "Jean-Philippe ARGAUD"

import math, numpy, scipy
from daCore.NumericObjects import ApplyBounds, ForceNumericBounds
from daCore.PlatformInfo import PlatformInfo
mpr = PlatformInfo().MachinePrecision()
mfp = PlatformInfo().MaximumPrecision()

# ==============================================================================
def c2ukf(selfA, Xb, Y, U, HO, EM, CM, R, B, Q):
    """
    Constrained (2UKF) Unscented Kalman Filter
    """
    if selfA._parameters["EstimationOf"] == "Parameters":
        selfA._parameters["StoreInternalVariables"] = True
    selfA._parameters["Bounds"] = ForceNumericBounds( selfA._parameters["Bounds"] )
    #
    L     = Xb.size
    Alpha = selfA._parameters["Alpha"]
    Beta  = selfA._parameters["Beta"]
    if selfA._parameters["Kappa"] == 0:
        if selfA._parameters["EstimationOf"] == "State":
            Kappa = 0
        elif selfA._parameters["EstimationOf"] == "Parameters":
            Kappa = 3 - L
    else:
        Kappa = selfA._parameters["Kappa"]
    Lambda = float( Alpha**2 ) * ( L + Kappa ) - L
    Gamma  = math.sqrt( L + Lambda )
    #
    Ww = []
    Ww.append( 0. )
    for i in range(2*L):
        Ww.append( 1. / (2.*(L + Lambda)) )
    #
    Wm = numpy.array( Ww )
    Wm[0] = Lambda / (L + Lambda)
    Wc = numpy.array( Ww )
    Wc[0] = Lambda / (L + Lambda) + (1. - Alpha**2 + Beta)
    #
    # Durée d'observation et tailles
    if hasattr(Y,"stepnumber"):
        duration = Y.stepnumber()
        __p = numpy.cumprod(Y.shape())[-1]
    else:
        duration = 2
        __p = numpy.array(Y).size
    #
    # Précalcul des inversions de B et R
    if selfA._parameters["StoreInternalVariables"] \
        or selfA._toStore("CostFunctionJ") \
        or selfA._toStore("CostFunctionJb") \
        or selfA._toStore("CostFunctionJo") \
        or selfA._toStore("CurrentOptimum") \
        or selfA._toStore("APosterioriCovariance"):
        BI = B.getI()
        RI = R.getI()
    #
    __n = Xb.size
    nbPreviousSteps  = len(selfA.StoredVariables["Analysis"])
    #
    if len(selfA.StoredVariables["Analysis"])==0 or not selfA._parameters["nextStep"]:
        Xn = Xb
        if hasattr(B,"asfullmatrix"):
            Pn = B.asfullmatrix(__n)
        else:
            Pn = B
        selfA.StoredVariables["CurrentIterationNumber"].store( len(selfA.StoredVariables["Analysis"]) )
        selfA.StoredVariables["Analysis"].store( Xb )
        if selfA._toStore("APosterioriCovariance"):
            selfA.StoredVariables["APosterioriCovariance"].store( Pn )
    elif selfA._parameters["nextStep"]:
        Xn = selfA._getInternalState("Xn")
        Pn = selfA._getInternalState("Pn")
    #
    if selfA._parameters["EstimationOf"] == "Parameters":
        XaMin            = Xn
        previousJMinimum = numpy.finfo(float).max
    #
    for step in range(duration-1):
        #
        if U is not None:
            if hasattr(U,"store") and len(U)>1:
                Un = numpy.ravel( U[step] ).reshape((-1,1))
            elif hasattr(U,"store") and len(U)==1:
                Un = numpy.ravel( U[0] ).reshape((-1,1))
            else:
                Un = numpy.ravel( U ).reshape((-1,1))
        else:
            Un = None
        #
        if CM is not None and "Tangent" in CM and U is not None:
            Cm = CM["Tangent"].asMatrix(Xn)
        else:
            Cm = None
        #
        Pndemi = numpy.real(scipy.linalg.sqrtm(Pn))
        Xnp = numpy.hstack([Xn, Xn+Gamma*Pndemi, Xn-Gamma*Pndemi])
        nbSpts = 2*Xn.size+1
        #
        if selfA._parameters["Bounds"] is not None and selfA._parameters["ConstrainedBy"] == "EstimateProjection":
            for point in range(nbSpts):
                Xnp[:,point] = ApplyBounds( Xnp[:,point], selfA._parameters["Bounds"] )
        #
        XEtnnp = []
        for point in range(nbSpts):
            if selfA._parameters["EstimationOf"] == "State":
                Mm = EM["Direct"].appliedControledFormTo
                XEtnnpi = numpy.asarray( Mm( (Xnp[:,point], Un) ) ).reshape((-1,1))
                if Cm is not None and Un is not None: # Attention : si Cm est aussi dans M, doublon !
                    Cm = Cm.reshape(Xn.size,Un.size) # ADAO & check shape
                    XEtnnpi = XEtnnpi + Cm @ Un
                if selfA._parameters["Bounds"] is not None and selfA._parameters["ConstrainedBy"] == "EstimateProjection":
                    XEtnnpi = ApplyBounds( XEtnnpi, selfA._parameters["Bounds"] )
            elif selfA._parameters["EstimationOf"] == "Parameters":
                # --- > Par principe, M = Id, Q = 0
                XEtnnpi = Xnp[:,point]
            XEtnnp.append( numpy.ravel(XEtnnpi).reshape((-1,1)) )
        XEtnnp = numpy.concatenate( XEtnnp, axis=1 )
        #
        Xncm = ( XEtnnp * Wm ).sum(axis=1)
        #
        if selfA._parameters["Bounds"] is not None and selfA._parameters["ConstrainedBy"] == "EstimateProjection":
            Xncm = ApplyBounds( Xncm, selfA._parameters["Bounds"] )
        #
        if selfA._parameters["EstimationOf"] == "State":        Pnm = Q
        elif selfA._parameters["EstimationOf"] == "Parameters": Pnm = 0.
        for point in range(nbSpts):
            Pnm += Wc[i] * ((XEtnnp[:,point]-Xncm).reshape((-1,1)) * (XEtnnp[:,point]-Xncm))
        #
        if selfA._parameters["EstimationOf"] == "Parameters" and selfA._parameters["Bounds"] is not None:
            Pnmdemi = selfA._parameters["Reconditioner"] * numpy.real(scipy.linalg.sqrtm(Pnm))
        else:
            Pnmdemi = numpy.real(scipy.linalg.sqrtm(Pnm))
        #
        Xnnp = numpy.hstack([Xncm.reshape((-1,1)), Xncm.reshape((-1,1))+Gamma*Pnmdemi, Xncm.reshape((-1,1))-Gamma*Pnmdemi])
        #
        if selfA._parameters["Bounds"] is not None and selfA._parameters["ConstrainedBy"] == "EstimateProjection":
            for point in range(nbSpts):
                Xnnp[:,point] = ApplyBounds( Xnnp[:,point], selfA._parameters["Bounds"] )
        #
        Hm = HO["Direct"].appliedControledFormTo
        Ynnp = []
        for point in range(nbSpts):
            if selfA._parameters["EstimationOf"] == "State":
                Ynnpi = Hm( (Xnnp[:,point], None) )
            elif selfA._parameters["EstimationOf"] == "Parameters":
                Ynnpi = Hm( (Xnnp[:,point], Un) )
            Ynnp.append( numpy.ravel(Ynnpi).reshape((-1,1)) )
        Ynnp = numpy.concatenate( Ynnp, axis=1 )
        #
        Yncm = ( Ynnp * Wm ).sum(axis=1)
        #
        Pyyn = R
        Pxyn = 0.
        for point in range(nbSpts):
            Pyyn += Wc[i] * ((Ynnp[:,point]-Yncm).reshape((-1,1)) * (Ynnp[:,point]-Yncm))
            Pxyn += Wc[i] * ((Xnnp[:,point]-Xncm).reshape((-1,1)) * (Ynnp[:,point]-Yncm))
        #
        if hasattr(Y,"store"):
            Ynpu = numpy.ravel( Y[step+1] ).reshape((__p,1))
        else:
            Ynpu = numpy.ravel( Y ).reshape((__p,1))
        _Innovation  = Ynpu - Yncm.reshape((-1,1))
        if selfA._parameters["EstimationOf"] == "Parameters":
            if Cm is not None and Un is not None: # Attention : si Cm est aussi dans H, doublon !
                _Innovation = _Innovation - Cm @ Un
        #
        Kn = Pxyn * Pyyn.I
        Xn = Xncm.reshape((-1,1)) + Kn * _Innovation
        Pn = Pnm - Kn * Pyyn * Kn.T
        #
        if selfA._parameters["Bounds"] is not None and selfA._parameters["ConstrainedBy"] == "EstimateProjection":
            Xn = ApplyBounds( Xn, selfA._parameters["Bounds"] )
        #
        Xa = Xn # Pointeurs
        #--------------------------
        selfA._setInternalState("Xn", Xn)
        selfA._setInternalState("Pn", Pn)
        #--------------------------
        #
        selfA.StoredVariables["CurrentIterationNumber"].store( len(selfA.StoredVariables["Analysis"]) )
        # ---> avec analysis
        selfA.StoredVariables["Analysis"].store( Xa )
        if selfA._toStore("SimulatedObservationAtCurrentAnalysis"):
            selfA.StoredVariables["SimulatedObservationAtCurrentAnalysis"].store( Hm((Xa, Un)) )
        if selfA._toStore("InnovationAtCurrentAnalysis"):
            selfA.StoredVariables["InnovationAtCurrentAnalysis"].store( _Innovation )
        # ---> avec current state
        if selfA._parameters["StoreInternalVariables"] \
            or selfA._toStore("CurrentState"):
            selfA.StoredVariables["CurrentState"].store( Xn )
        if selfA._toStore("ForecastState"):
            selfA.StoredVariables["ForecastState"].store( Xncm )
        if selfA._toStore("ForecastCovariance"):
            selfA.StoredVariables["ForecastCovariance"].store( Pnm )
        if selfA._toStore("BMA"):
            selfA.StoredVariables["BMA"].store( Xncm - Xa )
        if selfA._toStore("InnovationAtCurrentState"):
            selfA.StoredVariables["InnovationAtCurrentState"].store( _Innovation )
        if selfA._toStore("SimulatedObservationAtCurrentState") \
            or selfA._toStore("SimulatedObservationAtCurrentOptimum"):
            selfA.StoredVariables["SimulatedObservationAtCurrentState"].store( Yncm )
        # ---> autres
        if selfA._parameters["StoreInternalVariables"] \
            or selfA._toStore("CostFunctionJ") \
            or selfA._toStore("CostFunctionJb") \
            or selfA._toStore("CostFunctionJo") \
            or selfA._toStore("CurrentOptimum") \
            or selfA._toStore("APosterioriCovariance"):
            Jb  = float( 0.5 * (Xa - Xb).T * (BI * (Xa - Xb)) )
            Jo  = float( 0.5 * _Innovation.T * (RI * _Innovation) )
            J   = Jb + Jo
            selfA.StoredVariables["CostFunctionJb"].store( Jb )
            selfA.StoredVariables["CostFunctionJo"].store( Jo )
            selfA.StoredVariables["CostFunctionJ" ].store( J )
            #
            if selfA._toStore("IndexOfOptimum") \
                or selfA._toStore("CurrentOptimum") \
                or selfA._toStore("CostFunctionJAtCurrentOptimum") \
                or selfA._toStore("CostFunctionJbAtCurrentOptimum") \
                or selfA._toStore("CostFunctionJoAtCurrentOptimum") \
                or selfA._toStore("SimulatedObservationAtCurrentOptimum"):
                IndexMin = numpy.argmin( selfA.StoredVariables["CostFunctionJ"][nbPreviousSteps:] ) + nbPreviousSteps
            if selfA._toStore("IndexOfOptimum"):
                selfA.StoredVariables["IndexOfOptimum"].store( IndexMin )
            if selfA._toStore("CurrentOptimum"):
                selfA.StoredVariables["CurrentOptimum"].store( selfA.StoredVariables["Analysis"][IndexMin] )
            if selfA._toStore("SimulatedObservationAtCurrentOptimum"):
                selfA.StoredVariables["SimulatedObservationAtCurrentOptimum"].store( selfA.StoredVariables["SimulatedObservationAtCurrentAnalysis"][IndexMin] )
            if selfA._toStore("CostFunctionJbAtCurrentOptimum"):
                selfA.StoredVariables["CostFunctionJbAtCurrentOptimum"].store( selfA.StoredVariables["CostFunctionJb"][IndexMin] )
            if selfA._toStore("CostFunctionJoAtCurrentOptimum"):
                selfA.StoredVariables["CostFunctionJoAtCurrentOptimum"].store( selfA.StoredVariables["CostFunctionJo"][IndexMin] )
            if selfA._toStore("CostFunctionJAtCurrentOptimum"):
                selfA.StoredVariables["CostFunctionJAtCurrentOptimum" ].store( selfA.StoredVariables["CostFunctionJ" ][IndexMin] )
        if selfA._toStore("APosterioriCovariance"):
            selfA.StoredVariables["APosterioriCovariance"].store( Pn )
        if selfA._parameters["EstimationOf"] == "Parameters" \
            and J < previousJMinimum:
            previousJMinimum    = J
            XaMin               = Xa
            if selfA._toStore("APosterioriCovariance"):
                covarianceXaMin = selfA.StoredVariables["APosterioriCovariance"][-1]
    #
    # Stockage final supplémentaire de l'optimum en estimation de paramètres
    # ----------------------------------------------------------------------
    if selfA._parameters["EstimationOf"] == "Parameters":
        selfA.StoredVariables["CurrentIterationNumber"].store( len(selfA.StoredVariables["Analysis"]) )
        selfA.StoredVariables["Analysis"].store( XaMin )
        if selfA._toStore("APosterioriCovariance"):
            selfA.StoredVariables["APosterioriCovariance"].store( covarianceXaMin )
        if selfA._toStore("BMA"):
            selfA.StoredVariables["BMA"].store( numpy.ravel(Xb) - numpy.ravel(XaMin) )
    #
    return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

__doc__ = """
    Non Linear Least Squares
"""
__author__ = "Jean-Philippe ARGAUD"

import numpy, scipy, scipy.optimize, scipy.version

# ==============================================================================
def ecwnlls(selfA, Xb, Y, U, HO, CM, R, B, __storeState = False):
    """
    Correction
    """
    #
    # Initialisations
    # ---------------
    Hm = HO["Direct"].appliedTo
    Ha = HO["Adjoint"].appliedInXTo
    #
    if HO["AppliedInX"] is not None and "HXb" in HO["AppliedInX"]:
        HXb = numpy.asarray(Hm( Xb, HO["AppliedInX"]["HXb"] ))
    else:
        HXb = numpy.asarray(Hm( Xb ))
    HXb = HXb.reshape((-1,1))
    if Y.size != HXb.size:
        raise ValueError("The size %i of observations Y and %i of observed calculation H(X) are different, they have to be identical."%(Y.size,HXb.size))
    if max(Y.shape) != max(HXb.shape):
        raise ValueError("The shapes %s of observations Y and %s of observed calculation H(X) are different, they have to be identical."%(Y.shape,HXb.shape))
    #
    RI = R.getI()
    if selfA._parameters["Minimizer"] == "LM":
        RdemiI = R.choleskyI()
    #
    Xini = selfA._parameters["InitializationPoint"]
    #
    # Définition de la fonction-coût
    # ------------------------------
    def CostFunction(x):
        _X  = numpy.asarray(x).reshape((-1,1))
        if selfA._parameters["StoreInternalVariables"] or \
            selfA._toStore("CurrentState") or \
            selfA._toStore("CurrentOptimum"):
            selfA.StoredVariables["CurrentState"].store( _X )
        _HX = numpy.asarray(Hm( _X )).reshape((-1,1))
        _Innovation = Y - _HX
        if selfA._toStore("SimulatedObservationAtCurrentState") or \
            selfA._toStore("SimulatedObservationAtCurrentOptimum"):
            selfA.StoredVariables["SimulatedObservationAtCurrentState"].store( _HX )
        if selfA._toStore("InnovationAtCurrentState"):
            selfA.StoredVariables["InnovationAtCurrentState"].store( _Innovation )
        #
        Jb  = 0.
        Jo  = float( 0.5 * _Innovation.T * (RI * _Innovation) )
        J   = Jb + Jo
        #
        selfA.StoredVariables["CurrentIterationNumber"].store( len(selfA.StoredVariables["CostFunctionJ"]) )
        selfA.StoredVariables["CostFunctionJb"].store( Jb )
        selfA.StoredVariables["CostFunctionJo"].store( Jo )
        selfA.StoredVariables["CostFunctionJ" ].store( J )
        if selfA._toStore("IndexOfOptimum") or \
            selfA._toStore("CurrentOptimum") or \
            selfA._toStore("CostFunctionJAtCurrentOptimum") or \
            selfA._toStore("CostFunctionJbAtCurrentOptimum") or \
            selfA._toStore("CostFunctionJoAtCurrentOptimum") or \
            selfA._toStore("SimulatedObservationAtCurrentOptimum"):
            IndexMin = numpy.argmin( selfA.StoredVariables["CostFunctionJ"][nbPreviousSteps:] ) + nbPreviousSteps
        if selfA._toStore("IndexOfOptimum"):
            selfA.StoredVariables["IndexOfOptimum"].store( IndexMin )
        if selfA._toStore("CurrentOptimum"):
            selfA.StoredVariables["CurrentOptimum"].store( selfA.StoredVariables["CurrentState"][IndexMin] )
        if selfA._toStore("SimulatedObservationAtCurrentOptimum"):
            selfA.StoredVariables["SimulatedObservationAtCurrentOptimum"].store( selfA.StoredVariables["SimulatedObservationAtCurrentState"][IndexMin] )
        if selfA._toStore("CostFunctionJbAtCurrentOptimum"):
            selfA.StoredVariables["CostFunctionJbAtCurrentOptimum"].store( selfA.StoredVariables["CostFunctionJb"][IndexMin] )
        if selfA._toStore("CostFunctionJoAtCurrentOptimum"):
            selfA.StoredVariables["CostFunctionJoAtCurrentOptimum"].store( selfA.StoredVariables["CostFunctionJo"][IndexMin] )
        if selfA._toStore("CostFunctionJAtCurrentOptimum"):
            selfA.StoredVariables["CostFunctionJAtCurrentOptimum" ].store( selfA.StoredVariables["CostFunctionJ" ][IndexMin] )
        return J
    #
    def GradientOfCostFunction(x):
        _X      = numpy.asarray(x).reshape((-1,1))
        _HX     = numpy.asarray(Hm( _X )).reshape((-1,1))
        GradJb  = 0.
        GradJo  = - Ha( (_X, RI * (Y - _HX)) )
        GradJ   = numpy.ravel( GradJb ) + numpy.ravel( GradJo )
        return GradJ
    #
    def CostFunctionLM(x):
        _X  = numpy.ravel( x ).reshape((-1,1))
        _HX = Hm( _X ).reshape((-1,1))
        _Innovation = Y - _HX
        Jb  = 0.
        Jo  = float( 0.5 * _Innovation.T * (RI * _Innovation) )
        J   = Jb + Jo
        if selfA._parameters["StoreInternalVariables"] or \
            selfA._toStore("CurrentState"):
            selfA.StoredVariables["CurrentState"].store( _X )
        selfA.StoredVariables["CostFunctionJb"].store( Jb )
        selfA.StoredVariables["CostFunctionJo"].store( Jo )
        selfA.StoredVariables["CostFunctionJ" ].store( J )
        #
        return numpy.ravel( RdemiI*_Innovation )
    #
    def GradientOfCostFunctionLM(x):
        _X      = x.reshape((-1,1))
        return - RdemiI*HO["Tangent"].asMatrix( _X )
    #
    # Minimisation de la fonctionnelle
    # --------------------------------
    nbPreviousSteps = selfA.StoredVariables["CostFunctionJ"].stepnumber()
    #
    if selfA._parameters["Minimizer"] == "LBFGSB":
        if "0.19" <= scipy.version.version <= "1.4.1":
            import daAlgorithms.Atoms.lbfgsbhlt as optimiseur
        else:
            import scipy.optimize as optimiseur
        Minimum, J_optimal, Informations = optimiseur.fmin_l_bfgs_b(
            func        = CostFunction,
            x0          = Xini,
            fprime      = GradientOfCostFunction,
            args        = (),
            bounds      = selfA._parameters["Bounds"],
            maxfun      = selfA._parameters["MaximumNumberOfIterations"]-1,
            factr       = selfA._parameters["CostDecrementTolerance"]*1.e14,
            pgtol       = selfA._parameters["ProjectedGradientTolerance"],
            iprint      = selfA._parameters["optiprint"],
            )
        # nfeval = Informations['funcalls']
        # rc     = Informations['warnflag']
    elif selfA._parameters["Minimizer"] == "TNC":
        Minimum, nfeval, rc = scipy.optimize.fmin_tnc(
            func        = CostFunction,
            x0          = Xini,
            fprime      = GradientOfCostFunction,
            args        = (),
            bounds      = selfA._parameters["Bounds"],
            maxfun      = selfA._parameters["MaximumNumberOfIterations"],
            pgtol       = selfA._parameters["ProjectedGradientTolerance"],
            ftol        = selfA._parameters["CostDecrementTolerance"],
            messages    = selfA._parameters["optmessages"],
            )
    elif selfA._parameters["Minimizer"] == "CG":
        Minimum, fopt, nfeval, grad_calls, rc = scipy.optimize.fmin_cg(
            f           = CostFunction,
            x0          = Xini,
            fprime      = GradientOfCostFunction,
            args        = (),
            maxiter     = selfA._parameters["MaximumNumberOfIterations"],
            gtol        = selfA._parameters["GradientNormTolerance"],
            disp        = selfA._parameters["optdisp"],
            full_output = True,
            )
    elif selfA._parameters["Minimizer"] == "NCG":
        Minimum, fopt, nfeval, grad_calls, hcalls, rc = scipy.optimize.fmin_ncg(
            f           = CostFunction,
            x0          = Xini,
            fprime      = GradientOfCostFunction,
            args        = (),
            maxiter     = selfA._parameters["MaximumNumberOfIterations"],
            avextol     = selfA._parameters["CostDecrementTolerance"],
            disp        = selfA._parameters["optdisp"],
            full_output = True,
            )
    elif selfA._parameters["Minimizer"] == "BFGS":
        Minimum, fopt, gopt, Hopt, nfeval, grad_calls, rc = scipy.optimize.fmin_bfgs(
            f           = CostFunction,
            x0          = Xini,
            fprime      = GradientOfCostFunction,
            args        = (),
            maxiter     = selfA._parameters["MaximumNumberOfIterations"],
            gtol        = selfA._parameters["GradientNormTolerance"],
            disp        = selfA._parameters["optdisp"],
            full_output = True,
            )
    elif selfA._parameters["Minimizer"] == "LM":
        Minimum, cov_x, infodict, mesg, rc = scipy.optimize.leastsq(
            func        = CostFunctionLM,
            x0          = Xini,
            Dfun        = GradientOfCostFunctionLM,
            args        = (),
            ftol        = selfA._parameters["CostDecrementTolerance"],
            maxfev      = selfA._parameters["MaximumNumberOfIterations"],
            gtol        = selfA._parameters["GradientNormTolerance"],
            full_output = True,
            )
        # nfeval = infodict['nfev']
    else:
        raise ValueError("Error in minimizer name: %s is unkown"%selfA._parameters["Minimizer"])
    #
    IndexMin = numpy.argmin( selfA.StoredVariables["CostFunctionJ"][nbPreviousSteps:] ) + nbPreviousSteps
    #
    # Correction pour pallier a un bug de TNC sur le retour du Minimum
    # ----------------------------------------------------------------
    if selfA._parameters["StoreInternalVariables"] or selfA._toStore("CurrentState"):
        Minimum = selfA.StoredVariables["CurrentState"][IndexMin]
    #
    Xa = Minimum
    if __storeState: selfA._setInternalState("Xn", Xa)
    #--------------------------
    #
    selfA.StoredVariables["Analysis"].store( Xa )
    #
    if selfA._toStore("OMA") or \
        selfA._toStore("InnovationAtCurrentAnalysis") or \
        selfA._toStore("SimulatedObservationAtOptimum"):
        if selfA._toStore("SimulatedObservationAtCurrentState"):
            HXa = selfA.StoredVariables["SimulatedObservationAtCurrentState"][IndexMin]
        elif selfA._toStore("SimulatedObservationAtCurrentOptimum"):
            HXa = selfA.StoredVariables["SimulatedObservationAtCurrentOptimum"][-1]
        else:
            HXa = Hm( Xa )
        oma = Y - HXa.reshape((-1,1))
    #
    # Calculs et/ou stockages supplémentaires
    # ---------------------------------------
    if selfA._toStore("Innovation") or \
        selfA._toStore("OMB"):
        Innovation  = Y - HXb
    if selfA._toStore("Innovation"):
        selfA.StoredVariables["Innovation"].store( Innovation )
    if selfA._toStore("BMA"):
        selfA.StoredVariables["BMA"].store( numpy.ravel(Xb) - numpy.ravel(Xa) )
    if selfA._toStore("OMA"):
        selfA.StoredVariables["OMA"].store( oma )
    if selfA._toStore("InnovationAtCurrentAnalysis"):
        selfA.StoredVariables["InnovationAtCurrentAnalysis"].store( oma )
    if selfA._toStore("OMB"):
        selfA.StoredVariables["OMB"].store( Innovation )
    if selfA._toStore("SimulatedObservationAtBackground"):
        selfA.StoredVariables["SimulatedObservationAtBackground"].store( HXb )
    if selfA._toStore("SimulatedObservationAtOptimum"):
        selfA.StoredVariables["SimulatedObservationAtOptimum"].store( HXa )
    #
    return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

__doc__ = """
    Extended BLUE
"""
__author__ = "Jean-Philippe ARGAUD"

import logging, numpy
from daCore.NumericObjects import QuantilesEstimations
from daCore.PlatformInfo import PlatformInfo
mpr = PlatformInfo().MachinePrecision()

# ==============================================================================
def ecwexblue(selfA, Xb, Y, U, HO, CM, R, B, __storeState = False):
    """
    Correction
    """
    #
    # Initialisations
    # ---------------
    Hm = HO["Tangent"].asMatrix(Xb)
    Hm = Hm.reshape(Y.size,Xb.size) # ADAO & check shape
    Ha = HO["Adjoint"].asMatrix(Xb)
    Ha = Ha.reshape(Xb.size,Y.size) # ADAO & check shape
    H  = HO["Direct"].appliedTo
    #
    if HO["AppliedInX"] is not None and "HXb" in HO["AppliedInX"]:
        HXb = numpy.asarray(H( Xb, HO["AppliedInX"]["HXb"]))
    else:
        HXb = numpy.asarray(H( Xb ))
    HXb = HXb.reshape((-1,1))
    if Y.size != HXb.size:
        raise ValueError("The size %i of observations Y and %i of observed calculation H(X) are different, they have to be identical."%(Y.size,HXb.size))
    if max(Y.shape) != max(HXb.shape):
        raise ValueError("The shapes %s of observations Y and %s of observed calculation H(X) are different, they have to be identical."%(Y.shape,HXb.shape))
    #
    if selfA._parameters["StoreInternalVariables"] or \
        selfA._toStore("CostFunctionJ")  or selfA._toStore("CostFunctionJAtCurrentOptimum") or \
        selfA._toStore("CostFunctionJb") or selfA._toStore("CostFunctionJbAtCurrentOptimum") or \
        selfA._toStore("CostFunctionJo") or selfA._toStore("CostFunctionJoAtCurrentOptimum") or \
        selfA._toStore("MahalanobisConsistency") or \
        (Y.size >  Xb.size):
        if isinstance(B,numpy.ndarray):
            BI = numpy.linalg.inv(B)
        else:
            BI = B.getI()
        RI = R.getI()
    #
    Innovation  = Y - HXb
    if selfA._parameters["EstimationOf"] == "Parameters":
        if CM is not None and "Tangent" in CM and U is not None: # Attention : si Cm est aussi dans H, doublon !
            Cm = CM["Tangent"].asMatrix(Xb)
            Cm = Cm.reshape(Xb.size,U.size) # ADAO & check shape
            Innovation = Innovation - (Cm @ U).reshape((-1,1))
    #
    # Calcul de l'analyse
    # -------------------
    if Y.size <= Xb.size:
        _HNHt = numpy.dot(Hm, B @ Ha)
        _A = R + _HNHt
        _u = numpy.linalg.solve( _A , numpy.ravel(Innovation) )
        Xa = Xb + (B @ numpy.ravel(Ha @ _u)).reshape((-1,1))
    else:
        _HtRH = numpy.dot(Ha, RI @ Hm)
        _A = BI + _HtRH
        _u = numpy.linalg.solve( _A , numpy.ravel(numpy.dot(Ha, RI @ numpy.ravel(Innovation))) )
        Xa = Xb + _u.reshape((-1,1))
    #
    if __storeState: selfA._setInternalState("Xn", Xa)
    #--------------------------
    #
    selfA.StoredVariables["Analysis"].store( Xa )
    #
    # Calcul de la fonction coût
    # --------------------------
    if selfA._parameters["StoreInternalVariables"] or \
        selfA._toStore("CostFunctionJ")  or selfA._toStore("CostFunctionJAtCurrentOptimum") or \
        selfA._toStore("CostFunctionJb") or selfA._toStore("CostFunctionJbAtCurrentOptimum") or \
        selfA._toStore("CostFunctionJo") or selfA._toStore("CostFunctionJoAtCurrentOptimum") or \
        selfA._toStore("OMA") or \
        selfA._toStore("InnovationAtCurrentAnalysis") or \
        selfA._toStore("SigmaObs2") or \
        selfA._toStore("MahalanobisConsistency") or \
        selfA._toStore("SimulatedObservationAtCurrentOptimum") or \
        selfA._toStore("SimulatedObservationAtCurrentState") or \
        selfA._toStore("SimulatedObservationAtOptimum") or \
        selfA._toStore("SimulationQuantiles"):
        HXa = H( Xa )
        oma = Y - HXa.reshape((-1,1))
    if selfA._parameters["StoreInternalVariables"] or \
        selfA._toStore("CostFunctionJ")  or selfA._toStore("CostFunctionJAtCurrentOptimum") or \
        selfA._toStore("CostFunctionJb") or selfA._toStore("CostFunctionJbAtCurrentOptimum") or \
        selfA._toStore("CostFunctionJo") or selfA._toStore("CostFunctionJoAtCurrentOptimum") or \
        selfA._toStore("MahalanobisConsistency"):
        Jb  = float( 0.5 * (Xa - Xb).T @ (BI @ (Xa - Xb)) )
        Jo  = float( 0.5 * oma.T * (RI * oma) )
        J   = Jb + Jo
        selfA.StoredVariables["CostFunctionJb"].store( Jb )
        selfA.StoredVariables["CostFunctionJo"].store( Jo )
        selfA.StoredVariables["CostFunctionJ" ].store( J )
        selfA.StoredVariables["CostFunctionJbAtCurrentOptimum"].store( Jb )
        selfA.StoredVariables["CostFunctionJoAtCurrentOptimum"].store( Jo )
        selfA.StoredVariables["CostFunctionJAtCurrentOptimum" ].store( J )
    #
    # Calcul de la covariance d'analyse
    # ---------------------------------
    if selfA._toStore("APosterioriCovariance") or \
        selfA._toStore("SimulationQuantiles"):
        if   (Y.size <= Xb.size): K  = B * Ha * (R + numpy.dot(Hm, B * Ha)).I
        elif (Y.size >  Xb.size): K = (BI + numpy.dot(Ha, RI * Hm)).I * Ha * RI
        A = B - K * Hm * B
        A = (A + A.T) * 0.5 # Symétrie
        A = A + mpr*numpy.trace( A ) * numpy.identity(Xa.size) # Positivité
        if min(A.shape) != max(A.shape):
            raise ValueError("The %s a posteriori covariance matrix A is of shape %s, despites it has to be a squared matrix. There is an error in the observation operator, please check it."%(selfA._name,str(A.shape)))
        if (numpy.diag(A) < 0).any():
            raise ValueError("The %s a posteriori covariance matrix A has at least one negative value on its diagonal. There is an error in the observation operator, please check it."%(selfA._name,))
        if logging.getLogger().level < logging.WARNING: # La vérification n'a lieu qu'en debug
            try:
                numpy.linalg.cholesky( A )
            except:
                raise ValueError("The %s a posteriori covariance matrix A is not symmetric positive-definite. Please check your a priori covariances and your observation operator."%(selfA._name,))
        selfA.StoredVariables["APosterioriCovariance"].store( A )
    #
    # Calculs et/ou stockages supplémentaires
    # ---------------------------------------
    if selfA._parameters["StoreInternalVariables"] or selfA._toStore("CurrentState"):
        selfA.StoredVariables["CurrentState"].store( Xa )
    if selfA._toStore("CurrentOptimum"):
        selfA.StoredVariables["CurrentOptimum"].store( Xa )
    if selfA._toStore("Innovation"):
        selfA.StoredVariables["Innovation"].store( Innovation )
    if selfA._toStore("BMA"):
        selfA.StoredVariables["BMA"].store( numpy.ravel(Xb) - numpy.ravel(Xa) )
    if selfA._toStore("OMA"):
        selfA.StoredVariables["OMA"].store( oma )
    if selfA._toStore("InnovationAtCurrentAnalysis"):
        selfA.StoredVariables["InnovationAtCurrentAnalysis"].store( oma )
    if selfA._toStore("OMB"):
        selfA.StoredVariables["OMB"].store( Innovation )
    if selfA._toStore("SigmaObs2"):
        TraceR = R.trace(Y.size)
        selfA.StoredVariables["SigmaObs2"].store( float( Innovation.T @ oma ) / TraceR )
    if selfA._toStore("SigmaBck2"):
        selfA.StoredVariables["SigmaBck2"].store( float( (Innovation.T @ (Hm @ (numpy.ravel(Xa) - numpy.ravel(Xb))))/(Hm * (B * Hm.T)).trace() ) )
    if selfA._toStore("MahalanobisConsistency"):
        selfA.StoredVariables["MahalanobisConsistency"].store( float( 2.*J/Innovation.size ) )
    if selfA._toStore("SimulationQuantiles"):
        HtM  = HO["Tangent"].asMatrix(Xa)
        HtM  = HtM.reshape(Y.size,Xa.size) # ADAO & check shape
        QuantilesEstimations(selfA, A, Xa, HXa, H, HtM)
    if selfA._toStore("SimulatedObservationAtBackground"):
        selfA.StoredVariables["SimulatedObservationAtBackground"].store( HXb )
    if selfA._toStore("SimulatedObservationAtCurrentState"):
        selfA.StoredVariables["SimulatedObservationAtCurrentState"].store( HXa )
    if selfA._toStore("SimulatedObservationAtCurrentOptimum"):
        selfA.StoredVariables["SimulatedObservationAtCurrentOptimum"].store( HXa )
    if selfA._toStore("SimulatedObservationAtOptimum"):
        selfA.StoredVariables["SimulatedObservationAtOptimum"].store( HXa )
    #
    return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This file is part of SALOME ADAO module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#

import os
import time
import subprocess
import traceback
import SalomePyQt

import eficasSalome
from Ihm import CONNECTOR
from . import adaoGuiHelper
from . import adaoStudyEditor

class AdaoCase:

    def __init__(self):

        self.name = "not yet defined"           # Name of the case

        self.filename      = "not yet defined"  # Python filename generated by Eficas
        self.yacs_filename = "not yet defined"  # Yacs schema filename
        self.tui_filename  = "not yet defined"  # Yacs schema filename

        #~ self.salome_study_id = -1            # Study of the case
        self.salome_study_item = None           # Study item object

        self.eficas_editor = None               # Editor object from Eficas
        self.arbreOuvert = False

    def setEditor(self, editor):
        if editor is not self.eficas_editor:
            self.eficas_editor = editor
            # Connect to the jdc
            CONNECTOR.Connect(self.eficas_editor.jdc, "valid", self.editorValidEvent, ())

    # Rq on notera que l'on utilise isValid dans isOk
    #    et que isOk appelle editorValidEvent
    #    il n'y a pas de boucle infinie car isValid n'émet
    #    son signal que si l'état a changé
    def editorValidEvent(self):
        adaoStudyEditor.updateItem(self.salome_study_item, self) # self.salome_study_id, self.salome_study_item, self)
        adaoGuiHelper.refreshObjectBrowser()

    def isOk(self):
        if self.eficas_editor.jdc:
            return self.eficas_editor.jdc.isValid()
        return False

    def createYACSFile(self):
        rtn = ""
        if (self.filename == "" or self.filename == "not yet defined"):
            return "You need to save your case before exporting it."

        self.yacs_filename = self.filename[:self.filename.rfind(".")] + '.xml'
        yacs_filename_backup = self.filename[:self.filename.rfind(".")] + '.xml.back'
        if os.path.exists(self.yacs_filename):
            if os.path.exists(yacs_filename_backup):
                os.unlink(yacs_filename_backup)
            os.rename(self.yacs_filename, yacs_filename_backup)

        self.eficas_editor.modified = True
        self.eficas_editor.saveFile()
        filename = self.filename[:self.filename.rfind(".")] + '.py'
        retry = 0
        while (not os.path.exists(filename) and retry < 30):
            time.sleep(0.1)
            retry += 1
        if not os.path.exists(filename):
            msg =  "Cannot find the COMM/PY associated EFICAS/Python files for YACS\n"
            msg += "generation. Is your case correct? Try to close and re-open the\n"
            msg += "case with the ADAO/EFICAS editor."
            return msg

        if "ADAO_ENGINE_ROOT_DIR" not in os.environ:
            return "Please add ADAO_ENGINE_ROOT_DIR to your environment."

        adao_path = os.environ["ADAO_ENGINE_ROOT_DIR"]
        adao_exe = adao_path + "/bin/AdaoYacsSchemaCreator.py"
        args = ["python", adao_exe, filename, self.yacs_filename]
        p = subprocess.Popen(args)
        (stdoutdata, stderrdata) = p.communicate()
        if not os.path.exists(self.yacs_filename):
            msg  = "An error occurred during the execution of the ADAO YACS\n"
            msg += "Schema Creator. If SALOME GUI is launched by command\n"
            msg += "line, see error details in your terminal.\n"
            return msg
        return rtn

    def exportTUIFile(self):
        if (self.filename == "" or self.filename == "not yet defined"):
          return "You need to save your case before exporting it."

        self.tui_filename = self.filename[:self.filename.rfind(".")] + '_TUI.py'
        tui_filename_backup = self.filename[:self.filename.rfind(".")] + '_TUI.py.back'
        if os.path.exists(self.tui_filename):
            if os.path.exists(tui_filename_backup):
                os.unlink(tui_filename_backup)
            os.rename(self.tui_filename, tui_filename_backup)

        self.eficas_editor.modified = True
        self.eficas_editor.saveFile()
        #
        import adaoBuilder
        fullcase = adaoBuilder.New()
        rtn = fullcase.convert(FileNameFrom=self.filename, FormaterFrom="COM",
                               FileNameTo=self.tui_filename, FormaterTo="TUI",
                              )
        del fullcase
        #
        if not os.path.exists(self.tui_filename) or len(rtn) == 0:
            msg  = "An error occurred during the conversion of the COMM case\n"
            msg += "to the TUI one. If SALOME GUI is launched by command\n"
            msg += "line, see error details in your terminal.\n"
        else:
            msg  = "The current ADAO case has been successfully exported to\n"
            msg += "the file named %s.\n"%self.tui_filename
            msg += "Take the time to review it before executing it as a case.\n"
        return msg

    def exportCaseToYACS(self):
        rtn = ""
        rtn = self.createYACSFile()
        if rtn != "":
            return rtn

        try:
            import libYACS_Swig
            yacs_swig = libYACS_Swig.YACS_Swig()
            #~ yacs_swig.loadSchema(self.yacs_filename, 1, 1)
            yacs_swig.loadSchema(self.yacs_filename, True, True)
        except:
            msg =  "Please install YACS module, error was: \n"
            msg += traceback.format_exc()
            return msg
        return rtn

    def validationReportforJDC(self):
        rtn = "<i>Validation report is empty.</i>"
        if self.eficas_editor.jdc:
            rtn  = "Validation report for the selected ADAO case:\n\n"
            rtn += str( self.eficas_editor.jdc.report() )
        return rtn

    def showTreeAdaoCase(self):
        if self.eficas_editor:
            if self.arbreOuvert:
                self.eficas_editor.fermeArbre()
                self.arbreOuvert = False
            else:
                self.eficas_editor.ouvreArbre()
                self.arbreOuvert = True
        return self.arbreOuvert



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

__doc__ = """
    Stochastic EnKF
"""
__author__ = "Jean-Philippe ARGAUD"

import math, numpy
from daCore.NumericObjects import Apply3DVarRecentringOnEnsemble
from daCore.NumericObjects import CovarianceInflation
from daCore.NumericObjects import EnsembleErrorCovariance
from daCore.NumericObjects import EnsembleMean
from daCore.NumericObjects import EnsembleOfAnomalies
from daCore.NumericObjects import EnsembleOfBackgroundPerturbations
from daCore.NumericObjects import EnsembleOfCenteredPerturbations
from daCore.NumericObjects import EnsemblePerturbationWithGivenCovariance
from daCore.PlatformInfo import PlatformInfo
mpr = PlatformInfo().MachinePrecision()
mfp = PlatformInfo().MaximumPrecision()

# ==============================================================================
def senkf(selfA, Xb, Y, U, HO, EM, CM, R, B, Q,
    VariantM="KalmanFilterFormula16",
    Hybrid=None,
    ):
    """
    Stochastic EnKF
    """
    if selfA._parameters["EstimationOf"] == "Parameters":
        selfA._parameters["StoreInternalVariables"] = True
    #
    # Opérateurs
    H = HO["Direct"].appliedControledFormTo
    #
    if selfA._parameters["EstimationOf"] == "State":
        M = EM["Direct"].appliedControledFormTo
    #
    if CM is not None and "Tangent" in CM and U is not None:
        Cm = CM["Tangent"].asMatrix(Xb)
    else:
        Cm = None
    #
    # Durée d'observation et tailles
    if hasattr(Y,"stepnumber"):
        duration = Y.stepnumber()
        __p = numpy.cumprod(Y.shape())[-1]
    else:
        duration = 2
        __p = numpy.array(Y).size
    #
    # Précalcul des inversions de B et R
    if selfA._parameters["StoreInternalVariables"] \
        or selfA._toStore("CostFunctionJ") \
        or selfA._toStore("CostFunctionJb") \
        or selfA._toStore("CostFunctionJo") \
        or selfA._toStore("CurrentOptimum") \
        or selfA._toStore("APosterioriCovariance"):
        BI = B.getI()
        RI = R.getI()
    #
    __n = Xb.size
    __m = selfA._parameters["NumberOfMembers"]
    nbPreviousSteps  = len(selfA.StoredVariables["Analysis"])
    previousJMinimum = numpy.finfo(float).max
    #
    if hasattr(R,"asfullmatrix"): Rn = R.asfullmatrix(__p)
    else:                         Rn = R
    #
    if len(selfA.StoredVariables["Analysis"])==0 or not selfA._parameters["nextStep"]:
        if hasattr(B,"asfullmatrix"): Pn = B.asfullmatrix(__n)
        else:                         Pn = B
        Xn = EnsembleOfBackgroundPerturbations( Xb, Pn, __m )
        selfA.StoredVariables["Analysis"].store( Xb )
        if selfA._toStore("APosterioriCovariance"):
            selfA.StoredVariables["APosterioriCovariance"].store( Pn )
        selfA._setInternalState("seed", numpy.random.get_state())
    elif selfA._parameters["nextStep"]:
        Xn = selfA._getInternalState("Xn")
    #
    for step in range(duration-1):
        numpy.random.set_state(selfA._getInternalState("seed"))
        if hasattr(Y,"store"):
            Ynpu = numpy.ravel( Y[step+1] ).reshape((__p,1))
        else:
            Ynpu = numpy.ravel( Y ).reshape((__p,1))
        #
        if U is not None:
            if hasattr(U,"store") and len(U)>1:
                Un = numpy.ravel( U[step] ).reshape((-1,1))
            elif hasattr(U,"store") and len(U)==1:
                Un = numpy.ravel( U[0] ).reshape((-1,1))
            else:
                Un = numpy.ravel( U ).reshape((-1,1))
        else:
            Un = None
        #
        if selfA._parameters["InflationType"] == "MultiplicativeOnBackgroundAnomalies":
            Xn = CovarianceInflation( Xn,
                selfA._parameters["InflationType"],
                selfA._parameters["InflationFactor"],
                )
        #
        if selfA._parameters["EstimationOf"] == "State": # Forecast + Q and observation of forecast
            EMX = M( [(Xn[:,i], Un) for i in range(__m)],
                argsAsSerie = True,
                returnSerieAsArrayMatrix = True )
            Xn_predicted = EnsemblePerturbationWithGivenCovariance( EMX, Q )
            HX_predicted = H( [(Xn_predicted[:,i], Un) for i in range(__m)],
                argsAsSerie = True,
                returnSerieAsArrayMatrix = True )
            if Cm is not None and Un is not None: # Attention : si Cm est aussi dans M, doublon !
                Cm = Cm.reshape(__n,Un.size) # ADAO & check shape
                Xn_predicted = Xn_predicted + Cm @ Un
        elif selfA._parameters["EstimationOf"] == "Parameters": # Observation of forecast
            # --- > Par principe, M = Id, Q = 0
            Xn_predicted = EMX = Xn
            HX_predicted = H( [(Xn_predicted[:,i], Un) for i in range(__m)],
                argsAsSerie = True,
                returnSerieAsArrayMatrix = True )
        #
        # Mean of forecast and observation of forecast
        Xfm  = EnsembleMean( Xn_predicted )
        Hfm  = EnsembleMean( HX_predicted )
        #
        #--------------------------
        if VariantM == "KalmanFilterFormula05":
            PfHT, HPfHT = 0., 0.
            for i in range(__m):
                Exfi = Xn_predicted[:,i].reshape((__n,1)) - Xfm
                Eyfi = HX_predicted[:,i].reshape((__p,1)) - Hfm
                PfHT  += Exfi * Eyfi.T
                HPfHT += Eyfi * Eyfi.T
            PfHT  = (1./(__m-1)) * PfHT
            HPfHT = (1./(__m-1)) * HPfHT
            Kn     = PfHT * ( R + HPfHT ).I
            del PfHT, HPfHT
            #
            for i in range(__m):
                ri = numpy.random.multivariate_normal(numpy.zeros(__p), Rn)
                Xn[:,i] = numpy.ravel(Xn_predicted[:,i]) + Kn @ (numpy.ravel(Ynpu) + ri - HX_predicted[:,i])
        #--------------------------
        elif VariantM == "KalmanFilterFormula16":
            EpY   = EnsembleOfCenteredPerturbations(Ynpu, Rn, __m)
            EpYm  = EpY.mean(axis=1, dtype=mfp).astype('float').reshape((__p,1))
            #
            EaX   = EnsembleOfAnomalies( Xn_predicted ) / math.sqrt(__m-1)
            EaY = (HX_predicted - Hfm - EpY + EpYm) / math.sqrt(__m-1)
            #
            Kn = EaX @ EaY.T @ numpy.linalg.inv( EaY @ EaY.T)
            #
            for i in range(__m):
                Xn[:,i] = numpy.ravel(Xn_predicted[:,i]) + Kn @ (numpy.ravel(EpY[:,i]) - HX_predicted[:,i])
        #--------------------------
        else:
            raise ValueError("VariantM has to be chosen in the authorized methods list.")
        #
        if selfA._parameters["InflationType"] == "MultiplicativeOnAnalysisAnomalies":
            Xn = CovarianceInflation( Xn,
                selfA._parameters["InflationType"],
                selfA._parameters["InflationFactor"],
                )
        #
        if Hybrid == "E3DVAR":
            Xn = Apply3DVarRecentringOnEnsemble(Xn, EMX, Ynpu, HO, R, B, selfA._parameters)
        #
        Xa = EnsembleMean( Xn )
        #--------------------------
        selfA._setInternalState("Xn", Xn)
        selfA._setInternalState("seed", numpy.random.get_state())
        #--------------------------
        #
        if selfA._parameters["StoreInternalVariables"] \
            or selfA._toStore("CostFunctionJ") \
            or selfA._toStore("CostFunctionJb") \
            or selfA._toStore("CostFunctionJo") \
            or selfA._toStore("APosterioriCovariance") \
            or selfA._toStore("InnovationAtCurrentAnalysis") \
            or selfA._toStore("SimulatedObservationAtCurrentAnalysis") \
            or selfA._toStore("SimulatedObservationAtCurrentOptimum"):
            _HXa = numpy.ravel( H((Xa, Un)) ).reshape((-1,1))
            _Innovation = Ynpu - _HXa
        #
        selfA.StoredVariables["CurrentIterationNumber"].store( len(selfA.StoredVariables["Analysis"]) )
        # ---> avec analysis
        selfA.StoredVariables["Analysis"].store( Xa )
        if selfA._toStore("SimulatedObservationAtCurrentAnalysis"):
            selfA.StoredVariables["SimulatedObservationAtCurrentAnalysis"].store( _HXa )
        if selfA._toStore("InnovationAtCurrentAnalysis"):
            selfA.StoredVariables["InnovationAtCurrentAnalysis"].store( _Innovation )
        # ---> avec current state
        if selfA._parameters["StoreInternalVariables"] \
            or selfA._toStore("CurrentState"):
            selfA.StoredVariables["CurrentState"].store( Xn )
        if selfA._toStore("ForecastState"):
            selfA.StoredVariables["ForecastState"].store( EMX )
        if selfA._toStore("ForecastCovariance"):
            selfA.StoredVariables["ForecastCovariance"].store( EnsembleErrorCovariance(EMX) )
        if selfA._toStore("BMA"):
            selfA.StoredVariables["BMA"].store( EMX - Xa )
        if selfA._toStore("InnovationAtCurrentState"):
            selfA.StoredVariables["InnovationAtCurrentState"].store( - HX_predicted + Ynpu )
        if selfA._toStore("SimulatedObservationAtCurrentState") \
            or selfA._toStore("SimulatedObservationAtCurrentOptimum"):
            selfA.StoredVariables["SimulatedObservationAtCurrentState"].store( HX_predicted )
        # ---> autres
        if selfA._parameters["StoreInternalVariables"] \
            or selfA._toStore("CostFunctionJ") \
            or selfA._toStore("CostFunctionJb") \
            or selfA._toStore("CostFunctionJo") \
            or selfA._toStore("CurrentOptimum") \
            or selfA._toStore("APosterioriCovariance"):
            Jb  = float( 0.5 * (Xa - Xb).T * (BI * (Xa - Xb)) )
            Jo  = float( 0.5 * _Innovation.T * (RI * _Innovation) )
            J   = Jb + Jo
            selfA.StoredVariables["CostFunctionJb"].store( Jb )
            selfA.StoredVariables["CostFunctionJo"].store( Jo )
            selfA.StoredVariables["CostFunctionJ" ].store( J )
            #
            if selfA._toStore("IndexOfOptimum") \
                or selfA._toStore("CurrentOptimum") \
                or selfA._toStore("CostFunctionJAtCurrentOptimum") \
                or selfA._toStore("CostFunctionJbAtCurrentOptimum") \
                or selfA._toStore("CostFunctionJoAtCurrentOptimum") \
                or selfA._toStore("SimulatedObservationAtCurrentOptimum"):
                IndexMin = numpy.argmin( selfA.StoredVariables["CostFunctionJ"][nbPreviousSteps:] ) + nbPreviousSteps
            if selfA._toStore("IndexOfOptimum"):
                selfA.StoredVariables["IndexOfOptimum"].store( IndexMin )
            if selfA._toStore("CurrentOptimum"):
                selfA.StoredVariables["CurrentOptimum"].store( selfA.StoredVariables["Analysis"][IndexMin] )
            if selfA._toStore("SimulatedObservationAtCurrentOptimum"):
                selfA.StoredVariables["SimulatedObservationAtCurrentOptimum"].store( selfA.StoredVariables["SimulatedObservationAtCurrentAnalysis"][IndexMin] )
            if selfA._toStore("CostFunctionJbAtCurrentOptimum"):
                selfA.StoredVariables["CostFunctionJbAtCurrentOptimum"].store( selfA.StoredVariables["CostFunctionJb"][IndexMin] )
            if selfA._toStore("CostFunctionJoAtCurrentOptimum"):
                selfA.StoredVariables["CostFunctionJoAtCurrentOptimum"].store( selfA.StoredVariables["CostFunctionJo"][IndexMin] )
            if selfA._toStore("CostFunctionJAtCurrentOptimum"):
                selfA.StoredVariables["CostFunctionJAtCurrentOptimum" ].store( selfA.StoredVariables["CostFunctionJ" ][IndexMin] )
        if selfA._toStore("APosterioriCovariance"):
            selfA.StoredVariables["APosterioriCovariance"].store( EnsembleErrorCovariance(Xn) )
        if selfA._parameters["EstimationOf"] == "Parameters" \
            and J < previousJMinimum:
            previousJMinimum    = J
            XaMin               = Xa
            if selfA._toStore("APosterioriCovariance"):
                covarianceXaMin = selfA.StoredVariables["APosterioriCovariance"][-1]
        # ---> Pour les smoothers
        if selfA._toStore("CurrentEnsembleState"):
            selfA.StoredVariables["CurrentEnsembleState"].store( Xn )
    #
    # Stockage final supplémentaire de l'optimum en estimation de paramètres
    # ----------------------------------------------------------------------
    if selfA._parameters["EstimationOf"] == "Parameters":
        selfA.StoredVariables["CurrentIterationNumber"].store( len(selfA.StoredVariables["Analysis"]) )
        selfA.StoredVariables["Analysis"].store( XaMin )
        if selfA._toStore("APosterioriCovariance"):
            selfA.StoredVariables["APosterioriCovariance"].store( covarianceXaMin )
        if selfA._toStore("BMA"):
            selfA.StoredVariables["BMA"].store( numpy.ravel(Xb) - numpy.ravel(XaMin) )
    #
    return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This file is part of SALOME ADAO module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#

__author__="aribes/gboulant"

from daUtils.enumerate import Enumerate

from salome.kernel import studyedit

from . import adaoModuleHelper

#
# ==============================================================================
# Constant parameters and identifiers
# ==============================================================================
#
ADAO_ITEM_TYPES = Enumerate([
    "ADAO_CASE",
])

#
# ==============================================================================
# Function dedicated to the data management in the salome study
# ==============================================================================
#
# For developpers, note that the data structures used here are:
# - the SALOME study whose API is defined by SALOMEDS::Study
# - an item in a study whose API is defined by SALOMEDS:SObject
# - a study component, whose API is defined by SALOMEDS::SComponent
#   a SComponent is a reference in a study toward a SALOME component
#

def addInStudy(adaoCase): # salomeStudyId, adaoCase):
    """
    This function adds in the specified SALOME study a study entry that corresponds
    to the specified adao case. This study case is put in a folder under
    the ADAO component and is identified by the case name.
    """

    studyEditor = studyedit.getStudyEditor() # salomeStudyId)

    adaoRootEntry = studyEditor.findOrCreateComponent(
        moduleName    = adaoModuleHelper.componentName(),
        componentName = adaoModuleHelper.componentUserName(),
        icon          = adaoModuleHelper.modulePixmap())

    itemName  = adaoCase.name
    itemValue = adaoCase.filename
    itemType  = ADAO_ITEM_TYPES.ADAO_CASE

    icon = adaoModuleHelper.studyItemPixmapNOk()
    if adaoCase.isOk():
      icon = adaoModuleHelper.studyItemPixmapOk()

    salomeStudyItem = studyEditor.createItem(
        adaoRootEntry, itemName,
        comment = itemValue,
        typeId  = itemType,
        icon    = icon)

    return salomeStudyItem

def updateItem(salomeStudyItem, adaoCase): # salomeStudyId, salomeStudyItem, adaoCase):

    studyEditor = studyedit.getStudyEditor() # salomeStudyId)

    if salomeStudyItem.GetName()[:-2] != adaoCase.name:
      itemName  = adaoCase.name
      itemValue = adaoCase.filename
    else:
      itemName  = salomeStudyItem.GetName()
      itemValue = adaoCase.filename

    icon = adaoModuleHelper.studyItemPixmapNOk()
    if adaoCase.isOk():
      icon = adaoModuleHelper.studyItemPixmapOk()

    studyEditor.setItem(salomeStudyItem,
        name    = itemName,
        comment = itemValue,
        icon    = icon)

def removeItem(item): # salomeStudyId, item):
    """
    Remove the item from the specified study.
    """
    studyEditor = studyedit.getStudyEditor() # salomeStudyId)
    return studyEditor.removeItem(item,True)


def isValidAdaoCaseItem(item): # salomeStudyId,item):
    """
    Return true if the specified item corresponds to a valid adaoCase
    """
    if item is None:
        return False

    studyEditor = studyedit.getStudyEditor() # salomeStudyId)
    itemType = studyEditor.getTypeId(item)
    if itemType != ADAO_ITEM_TYPES.ADAO_CASE:
        return False

    return True



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

__doc__ = """
    4DVAR
"""
__author__ = "Jean-Philippe ARGAUD"

import numpy, scipy, scipy.optimize, scipy.version
from daCore.NumericObjects import ForceNumericBounds, ApplyBounds
from daCore.PlatformInfo import PlatformInfo
mpr = PlatformInfo().MachinePrecision()
mfp = PlatformInfo().MaximumPrecision()

# ==============================================================================
def std4dvar(selfA, Xb, Y, U, HO, EM, CM, R, B, Q):
    """
    Correction
    """
    #
    # Initialisations
    # ---------------
    #
    # Opérateurs
    Hm = HO["Direct"].appliedControledFormTo
    Mm = EM["Direct"].appliedControledFormTo
    #
    if CM is not None and "Tangent" in CM and U is not None:
        Cm = CM["Tangent"].asMatrix(Xb)
    else:
        Cm = None
    #
    def Un(_step):
        if U is not None:
            if hasattr(U,"store") and 1<=_step<len(U) :
                _Un = numpy.ravel( U[_step] ).reshape((-1,1))
            elif hasattr(U,"store") and len(U)==1:
                _Un = numpy.ravel( U[0] ).reshape((-1,1))
            else:
                _Un = numpy.ravel( U ).reshape((-1,1))
        else:
            _Un = None
        return _Un
    def CmUn(_xn,_un):
        if Cm is not None and _un is not None: # Attention : si Cm est aussi dans M, doublon !
            _Cm   = Cm.reshape(_xn.size,_un.size) # ADAO & check shape
            _CmUn = (_Cm @ _un).reshape((-1,1))
        else:
            _CmUn = 0.
        return _CmUn
    #
    # Remarque : les observations sont exploitées à partir du pas de temps
    # numéro 1, et sont utilisées dans Yo comme rangées selon ces indices.
    # Donc le pas 0 n'est pas utilisé puisque la première étape commence
    # avec l'observation du pas 1.
    #
    # Nombre de pas identique au nombre de pas d'observations
    if hasattr(Y,"stepnumber"):
        duration = Y.stepnumber()
    else:
        duration = 2
    #
    # Précalcul des inversions de B et R
    BI = B.getI()
    RI = R.getI()
    #
    # Point de démarrage de l'optimisation
    Xini = selfA._parameters["InitializationPoint"]
    #
    # Définition de la fonction-coût
    # ------------------------------
    selfA.DirectCalculation = [None,] # Le pas 0 n'est pas observé
    selfA.DirectInnovation  = [None,] # Le pas 0 n'est pas observé
    def CostFunction(x):
        _X  = numpy.asarray(x).reshape((-1,1))
        if selfA._parameters["StoreInternalVariables"] or \
            selfA._toStore("CurrentState") or \
            selfA._toStore("CurrentOptimum"):
            selfA.StoredVariables["CurrentState"].store( _X )
        Jb  = float( 0.5 * (_X - Xb).T * (BI * (_X - Xb)) )
        selfA.DirectCalculation = [None,]
        selfA.DirectInnovation  = [None,]
        Jo  = 0.
        _Xn = _X
        for step in range(0,duration-1):
            if hasattr(Y,"store"):
                _Ynpu = numpy.ravel( Y[step+1] ).reshape((-1,1))
            else:
                _Ynpu = numpy.ravel( Y ).reshape((-1,1))
            _Un = Un(step)
            #
            # Etape d'évolution
            if selfA._parameters["EstimationOf"] == "State":
                _Xn = Mm( (_Xn, _Un) ).reshape((-1,1)) + CmUn(_Xn, _Un)
            elif selfA._parameters["EstimationOf"] == "Parameters":
                pass
            #
            if selfA._parameters["Bounds"] is not None and selfA._parameters["ConstrainedBy"] == "EstimateProjection":
                _Xn = ApplyBounds( _Xn, ForceNumericBounds(selfA._parameters["Bounds"]) )
            #
            # Etape de différence aux observations
            if selfA._parameters["EstimationOf"] == "State":
                _YmHMX = _Ynpu - numpy.ravel( Hm( (_Xn, None) ) ).reshape((-1,1))
            elif selfA._parameters["EstimationOf"] == "Parameters":
                _YmHMX = _Ynpu - numpy.ravel( Hm( (_Xn, _Un) ) ).reshape((-1,1)) - CmUn(_Xn, _Un)
            #
            # Stockage de l'état
            selfA.DirectCalculation.append( _Xn )
            selfA.DirectInnovation.append( _YmHMX )
            #
            # Ajout dans la fonctionnelle d'observation
            Jo = Jo + 0.5 * float( _YmHMX.T * (RI * _YmHMX) )
        J = Jb + Jo
        #
        selfA.StoredVariables["CurrentIterationNumber"].store( len(selfA.StoredVariables["CostFunctionJ"]) )
        selfA.StoredVariables["CostFunctionJb"].store( Jb )
        selfA.StoredVariables["CostFunctionJo"].store( Jo )
        selfA.StoredVariables["CostFunctionJ" ].store( J )
        if selfA._toStore("IndexOfOptimum") or \
            selfA._toStore("CurrentOptimum") or \
            selfA._toStore("CostFunctionJAtCurrentOptimum") or \
            selfA._toStore("CostFunctionJbAtCurrentOptimum") or \
            selfA._toStore("CostFunctionJoAtCurrentOptimum"):
            IndexMin = numpy.argmin( selfA.StoredVariables["CostFunctionJ"][nbPreviousSteps:] ) + nbPreviousSteps
        if selfA._toStore("IndexOfOptimum"):
            selfA.StoredVariables["IndexOfOptimum"].store( IndexMin )
        if selfA._toStore("CurrentOptimum"):
            selfA.StoredVariables["CurrentOptimum"].store( selfA.StoredVariables["CurrentState"][IndexMin] )
        if selfA._toStore("CostFunctionJAtCurrentOptimum"):
            selfA.StoredVariables["CostFunctionJAtCurrentOptimum" ].store( selfA.StoredVariables["CostFunctionJ" ][IndexMin] )
        if selfA._toStore("CostFunctionJbAtCurrentOptimum"):
            selfA.StoredVariables["CostFunctionJbAtCurrentOptimum"].store( selfA.StoredVariables["CostFunctionJb"][IndexMin] )
        if selfA._toStore("CostFunctionJoAtCurrentOptimum"):
            selfA.StoredVariables["CostFunctionJoAtCurrentOptimum"].store( selfA.StoredVariables["CostFunctionJo"][IndexMin] )
        return J
    #
    def GradientOfCostFunction(x):
        _X      = numpy.asarray(x).reshape((-1,1))
        GradJb  = BI * (_X - Xb)
        GradJo  = 0.
        for step in range(duration-1,0,-1):
            # Étape de récupération du dernier stockage de l'évolution
            _Xn = selfA.DirectCalculation.pop()
            # Étape de récupération du dernier stockage de l'innovation
            _YmHMX = selfA.DirectInnovation.pop()
            # Calcul des adjoints
            Ha = HO["Adjoint"].asMatrix(ValueForMethodForm = _Xn)
            Ha = Ha.reshape(_Xn.size,_YmHMX.size) # ADAO & check shape
            Ma = EM["Adjoint"].asMatrix(ValueForMethodForm = _Xn)
            Ma = Ma.reshape(_Xn.size,_Xn.size) # ADAO & check shape
            # Calcul du gradient par état adjoint
            GradJo = GradJo + Ha * (RI * _YmHMX) # Équivaut pour Ha linéaire à : Ha( (_Xn, RI * _YmHMX) )
            GradJo = Ma * GradJo                 # Équivaut pour Ma linéaire à : Ma( (_Xn, GradJo) )
        GradJ = numpy.ravel( GradJb ) - numpy.ravel( GradJo )
        return GradJ
    #
    # Minimisation de la fonctionnelle
    # --------------------------------
    nbPreviousSteps = selfA.StoredVariables["CostFunctionJ"].stepnumber()
    #
    if selfA._parameters["Minimizer"] == "LBFGSB":
        if "0.19" <= scipy.version.version <= "1.4.1":
            import daAlgorithms.Atoms.lbfgsbhlt as optimiseur
        else:
            import scipy.optimize as optimiseur
        Minimum, J_optimal, Informations = optimiseur.fmin_l_bfgs_b(
            func        = CostFunction,
            x0          = Xini,
            fprime      = GradientOfCostFunction,
            args        = (),
            bounds      = selfA._parameters["Bounds"],
            maxfun      = selfA._parameters["MaximumNumberOfIterations"]-1,
            factr       = selfA._parameters["CostDecrementTolerance"]*1.e14,
            pgtol       = selfA._parameters["ProjectedGradientTolerance"],
            iprint      = selfA._parameters["optiprint"],
            )
        # nfeval = Informations['funcalls']
        # rc     = Informations['warnflag']
    elif selfA._parameters["Minimizer"] == "TNC":
        Minimum, nfeval, rc = scipy.optimize.fmin_tnc(
            func        = CostFunction,
            x0          = Xini,
            fprime      = GradientOfCostFunction,
            args        = (),
            bounds      = selfA._parameters["Bounds"],
            maxfun      = selfA._parameters["MaximumNumberOfIterations"],
            pgtol       = selfA._parameters["ProjectedGradientTolerance"],
            ftol        = selfA._parameters["CostDecrementTolerance"],
            messages    = selfA._parameters["optmessages"],
            )
    elif selfA._parameters["Minimizer"] == "CG":
        Minimum, fopt, nfeval, grad_calls, rc = scipy.optimize.fmin_cg(
            f           = CostFunction,
            x0          = Xini,
            fprime      = GradientOfCostFunction,
            args        = (),
            maxiter     = selfA._parameters["MaximumNumberOfIterations"],
            gtol        = selfA._parameters["GradientNormTolerance"],
            disp        = selfA._parameters["optdisp"],
            full_output = True,
            )
    elif selfA._parameters["Minimizer"] == "NCG":
        Minimum, fopt, nfeval, grad_calls, hcalls, rc = scipy.optimize.fmin_ncg(
            f           = CostFunction,
            x0          = Xini,
            fprime      = GradientOfCostFunction,
            args        = (),
            maxiter     = selfA._parameters["MaximumNumberOfIterations"],
            avextol     = selfA._parameters["CostDecrementTolerance"],
            disp        = selfA._parameters["optdisp"],
            full_output = True,
            )
    elif selfA._parameters["Minimizer"] == "BFGS":
        Minimum, fopt, gopt, Hopt, nfeval, grad_calls, rc = scipy.optimize.fmin_bfgs(
            f           = CostFunction,
            x0          = Xini,
            fprime      = GradientOfCostFunction,
            args        = (),
            maxiter     = selfA._parameters["MaximumNumberOfIterations"],
            gtol        = selfA._parameters["GradientNormTolerance"],
            disp        = selfA._parameters["optdisp"],
            full_output = True,
            )
    else:
        raise ValueError("Error in minimizer name: %s is unkown"%selfA._parameters["Minimizer"])
    #
    IndexMin = numpy.argmin( selfA.StoredVariables["CostFunctionJ"][nbPreviousSteps:] ) + nbPreviousSteps
    #
    # Correction pour pallier a un bug de TNC sur le retour du Minimum
    # ----------------------------------------------------------------
    if selfA._parameters["StoreInternalVariables"] or selfA._toStore("CurrentState"):
        Minimum = selfA.StoredVariables["CurrentState"][IndexMin]
    #
    # Obtention de l'analyse
    # ----------------------
    Xa = Minimum
    #
    selfA.StoredVariables["Analysis"].store( Xa )
    #
    # Calculs et/ou stockages supplémentaires
    # ---------------------------------------
    if selfA._toStore("BMA"):
        selfA.StoredVariables["BMA"].store( numpy.ravel(Xb) - numpy.ravel(Xa) )
    #
    return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

__doc__ = """
    Linear Least Squares
"""
__author__ = "Jean-Philippe ARGAUD"

# ==============================================================================
def ecwlls(selfA, Xb, Y, U, HO, CM, R, B, __storeState = False):
    """
    Correction
    """
    #
    # Initialisations
    # ---------------
    Hm = HO["Tangent"].asMatrix(Xb)
    Hm = Hm.reshape(Y.size,-1) # ADAO & check shape
    Ha = HO["Adjoint"].asMatrix(Xb)
    Ha = Ha.reshape(-1,Y.size) # ADAO & check shape
    #
    if R is None:
        RI = 1.
    else:
        RI = R.getI()
    #
    # Calcul de l'analyse
    # -------------------
    K = (Ha * (RI * Hm)).I * Ha * RI
    Xa =  K * Y
    #
    if __storeState: selfA._setInternalState("Xn", Xa)
    #--------------------------
    #
    selfA.StoredVariables["Analysis"].store( Xa )
    #
    # Calcul de la fonction coût
    # --------------------------
    if selfA._parameters["StoreInternalVariables"] or \
        selfA._toStore("CostFunctionJ")  or selfA._toStore("CostFunctionJAtCurrentOptimum") or \
        selfA._toStore("CostFunctionJb") or selfA._toStore("CostFunctionJbAtCurrentOptimum") or \
        selfA._toStore("CostFunctionJo") or selfA._toStore("CostFunctionJoAtCurrentOptimum") or \
        selfA._toStore("OMA") or \
        selfA._toStore("InnovationAtCurrentAnalysis") or \
        selfA._toStore("SimulatedObservationAtCurrentOptimum") or \
        selfA._toStore("SimulatedObservationAtCurrentState") or \
        selfA._toStore("SimulatedObservationAtOptimum"):
        HXa = Hm @ Xa
        oma = Y - HXa.reshape((-1,1))
    if selfA._parameters["StoreInternalVariables"] or \
        selfA._toStore("CostFunctionJ")  or selfA._toStore("CostFunctionJAtCurrentOptimum") or \
        selfA._toStore("CostFunctionJb") or selfA._toStore("CostFunctionJbAtCurrentOptimum") or \
        selfA._toStore("CostFunctionJo") or selfA._toStore("CostFunctionJoAtCurrentOptimum"):
        Jb  = 0.
        Jo  = float( 0.5 * oma.T * (RI * oma) )
        J   = Jb + Jo
        selfA.StoredVariables["CostFunctionJb"].store( Jb )
        selfA.StoredVariables["CostFunctionJo"].store( Jo )
        selfA.StoredVariables["CostFunctionJ" ].store( J )
        selfA.StoredVariables["CostFunctionJbAtCurrentOptimum"].store( Jb )
        selfA.StoredVariables["CostFunctionJoAtCurrentOptimum"].store( Jo )
        selfA.StoredVariables["CostFunctionJAtCurrentOptimum" ].store( J )
    #
    # Calculs et/ou stockages supplémentaires
    # ---------------------------------------
    if selfA._parameters["StoreInternalVariables"] or selfA._toStore("CurrentState"):
        selfA.StoredVariables["CurrentState"].store( Xa )
    if selfA._toStore("CurrentOptimum"):
        selfA.StoredVariables["CurrentOptimum"].store( Xa )
    if selfA._toStore("OMA"):
        selfA.StoredVariables["OMA"].store( oma )
    if selfA._toStore("InnovationAtCurrentAnalysis"):
        selfA.StoredVariables["InnovationAtCurrentAnalysis"].store( oma )
    if selfA._toStore("SimulatedObservationAtCurrentState"):
        selfA.StoredVariables["SimulatedObservationAtCurrentState"].store( HXa )
    if selfA._toStore("SimulatedObservationAtCurrentOptimum"):
        selfA.StoredVariables["SimulatedObservationAtCurrentOptimum"].store( HXa )
    if selfA._toStore("SimulatedObservationAtOptimum"):
        selfA.StoredVariables["SimulatedObservationAtOptimum"].store( HXa )
    #
    return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This file is part of SALOME ADAO module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

import os, sys

# print "import des prefs de Adao"
#
# Configuration de Eficas
# =======================
#
# INSTALLDIR, repIni est obligatoire
INSTALLDIR = ""
sys.path.insert(0,INSTALLDIR)
#
# Positionnee a repin au debut, mise a jour dans configuration
repIni=os.path.dirname(os.path.abspath(__file__))
#
# Sert comme directory initial des QFileDialog
initialdir = os.getcwd()
#
# Traductions et codages
#
# Codage des strings qui accepte les accents (en remplacement de 'ascii')
# encoding='utf-8'
#
# ===== Specifique : pour la 8_3_0 et suivantes 8_x ! Enlever pour les 9_x ! JPA
if sys.version.split()[0] < '3':
    # ===== Specifique : redefinition de l'encoding avant le Python 3 ! JPA
    reload(sys)
    sys.setdefaultencoding('utf-8')
    # ===== Specifique : pour les suivantes 8_x ! Enlever pour les 9_x ! JPA
    # 19/05/2017 : Redefinition du hook de "sys" present dans <GUI/src/SUITApp/SUITApp_init_python.cxx>
    def _custom_except_hook(exc_type, exc_value, exc_traceback):
        import sys
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        pass
    sys.excepthook = _custom_except_hook
    del _custom_except_hook, sys
# ===== Fin specifique ===================================================== JPA
#
# Indique la langue du catalogue utilisee pour les chaines d'aide : fr ou ang
# lang = 'fr'
# Traduction des labels de boutons ou autres
lookfor = os.path.abspath(os.path.join(os.path.dirname(__file__),"../resources"))
if os.path.exists(lookfor):
    translatorFichier = os.path.join(lookfor, "adao")
else:
    translatorFichier = os.environ["ADAO_ENGINE_ROOT_DIR"] + "/share/resources/adao/adao" # Ce nom sera complete par EFICAS avec _<LANG>.qm
#
# Pilotage des sous-fenetres d'EFICAS
closeAutreCommande = True
closeFrameRechercheCommande = True
closeFrameRechercheCommandeSurPageDesCommandes = True
closeEntete = True
closeArbre = True
taille = 800
nombreDeBoutonParLigne = 2




# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

__doc__ = """
    Extended Kalman Filter
"""
__author__ = "Jean-Philippe ARGAUD"

import numpy
from daCore.PlatformInfo import PlatformInfo
mpr = PlatformInfo().MachinePrecision()
mfp = PlatformInfo().MaximumPrecision()

# ==============================================================================
def exkf(selfA, Xb, Y, U, HO, EM, CM, R, B, Q):
    """
    Extended Kalman Filter
    """
    if selfA._parameters["EstimationOf"] == "Parameters":
        selfA._parameters["StoreInternalVariables"] = True
    #
    # Durée d'observation et tailles
    if hasattr(Y,"stepnumber"):
        duration = Y.stepnumber()
        __p = numpy.cumprod(Y.shape())[-1]
    else:
        duration = 2
        __p = numpy.array(Y).size
    __n = Xb.size
    #
    # Précalcul des inversions de B et R
    if selfA._parameters["StoreInternalVariables"] or \
        selfA._toStore("CostFunctionJ")  or selfA._toStore("CostFunctionJAtCurrentOptimum") or \
        selfA._toStore("CostFunctionJb") or selfA._toStore("CostFunctionJbAtCurrentOptimum") or \
        selfA._toStore("CostFunctionJo") or selfA._toStore("CostFunctionJoAtCurrentOptimum") or \
        selfA._toStore("CurrentOptimum") or selfA._toStore("APosterioriCovariance") or \
        (__p > __n):
        if isinstance(B,numpy.ndarray):
            BI = numpy.linalg.inv(B)
        else:
            BI = B.getI()
        RI = R.getI()
    #
    nbPreviousSteps  = len(selfA.StoredVariables["Analysis"])
    #
    if len(selfA.StoredVariables["Analysis"])==0 or not selfA._parameters["nextStep"]:
        Xn = Xb
        Pn = B
        selfA.StoredVariables["CurrentIterationNumber"].store( len(selfA.StoredVariables["Analysis"]) )
        selfA.StoredVariables["Analysis"].store( Xb )
        if selfA._toStore("APosterioriCovariance"):
            if hasattr(B,"asfullmatrix"):
                selfA.StoredVariables["APosterioriCovariance"].store( B.asfullmatrix(__n) )
            else:
                selfA.StoredVariables["APosterioriCovariance"].store( B )
        selfA._setInternalState("seed", numpy.random.get_state())
    elif selfA._parameters["nextStep"]:
        Xn = selfA._getInternalState("Xn")
        Pn = selfA._getInternalState("Pn")
    if hasattr(Pn,"asfullmatrix"):
        Pn = Pn.asfullmatrix(Xn.size)
    #
    if selfA._parameters["EstimationOf"] == "Parameters":
        XaMin            = Xn
        previousJMinimum = numpy.finfo(float).max
    #
    for step in range(duration-1):
        #
        if U is not None:
            if hasattr(U,"store") and len(U)>1:
                Un = numpy.ravel( U[step] ).reshape((-1,1))
            elif hasattr(U,"store") and len(U)==1:
                Un = numpy.ravel( U[0] ).reshape((-1,1))
            else:
                Un = numpy.ravel( U ).reshape((-1,1))
        else:
            Un = None
        #
        if selfA._parameters["EstimationOf"] == "State": # Forecast + Q and observation of forecast
            Mt = EM["Tangent"].asMatrix(Xn)
            Mt = Mt.reshape(Xn.size,Xn.size) # ADAO & check shape
            Ma = EM["Adjoint"].asMatrix(Xn)
            Ma = Ma.reshape(Xn.size,Xn.size) # ADAO & check shape
            M  = EM["Direct"].appliedControledFormTo
            Xn_predicted = numpy.ravel( M( (Xn, Un) ) ).reshape((__n,1))
            if CM is not None and "Tangent" in CM and Un is not None: # Attention : si Cm est aussi dans M, doublon !
                Cm = CM["Tangent"].asMatrix(Xn_predicted)
                Cm = Cm.reshape(__n,Un.size) # ADAO & check shape
                Xn_predicted = Xn_predicted + Cm @ Un
            Pn_predicted = Q + Mt @ (Pn @ Ma)
        elif selfA._parameters["EstimationOf"] == "Parameters": # Observation of forecast
            # --- > Par principe, M = Id, Q = 0
            Xn_predicted = Xn
            Pn_predicted = Pn
        #
        if hasattr(Y,"store"):
            Ynpu = numpy.ravel( Y[step+1] ).reshape((__p,1))
        else:
            Ynpu = numpy.ravel( Y ).reshape((__p,1))
        #
        Ht = HO["Tangent"].asMatrix(Xn_predicted)
        Ht = Ht.reshape(Ynpu.size,Xn.size) # ADAO & check shape
        Ha = HO["Adjoint"].asMatrix(Xn_predicted)
        Ha = Ha.reshape(Xn.size,Ynpu.size) # ADAO & check shape
        H  = HO["Direct"].appliedControledFormTo
        #
        if selfA._parameters["EstimationOf"] == "State":
            HX_predicted = numpy.ravel( H( (Xn_predicted, None) ) ).reshape((__p,1))
            _Innovation  = Ynpu - HX_predicted
        elif selfA._parameters["EstimationOf"] == "Parameters":
            HX_predicted = numpy.ravel( H( (Xn_predicted, Un) ) ).reshape((__p,1))
            _Innovation  = Ynpu - HX_predicted
            if CM is not None and "Tangent" in CM and Un is not None: # Attention : si Cm est aussi dans H, doublon !
                Cm = CM["Tangent"].asMatrix(Xn_predicted)
                Cm = Cm.reshape(__n,Un.size) # ADAO & check shape
                _Innovation = _Innovation - Cm @ Un
        #
        if Ynpu.size <= Xn.size:
            _HNHt = numpy.dot(Ht, Pn_predicted @ Ha)
            _A = R + _HNHt
            _u = numpy.linalg.solve( _A , _Innovation )
            Xn = Xn_predicted + (Pn_predicted @ (Ha @ _u)).reshape((-1,1))
            Kn = Pn_predicted @ (Ha @ numpy.linalg.inv(_A))
        else:
            _HtRH = numpy.dot(Ha, RI @ Ht)
            _A = numpy.linalg.inv(Pn_predicted) + _HtRH
            _u = numpy.linalg.solve( _A , numpy.dot(Ha, RI @ _Innovation) )
            Xn = Xn_predicted + _u.reshape((-1,1))
            Kn = numpy.linalg.inv(_A) @ (Ha @ RI.asfullmatrix(Ynpu.size))
        #
        Pn = Pn_predicted - Kn @ (Ht @ Pn_predicted)
        Pn = (Pn + Pn.T) * 0.5 # Symétrie
        Pn = Pn + mpr*numpy.trace( Pn ) * numpy.identity(Xn.size) # Positivité
        #
        Xa = Xn # Pointeurs
        #--------------------------
        selfA._setInternalState("Xn", Xn)
        selfA._setInternalState("Pn", Pn)
        #--------------------------
        #
        selfA.StoredVariables["CurrentIterationNumber"].store( len(selfA.StoredVariables["Analysis"]) )
        # ---> avec analysis
        selfA.StoredVariables["Analysis"].store( Xa )
        if selfA._toStore("SimulatedObservationAtCurrentAnalysis"):
            selfA.StoredVariables["SimulatedObservationAtCurrentAnalysis"].store( H((Xa, Un)) )
        if selfA._toStore("InnovationAtCurrentAnalysis"):
            selfA.StoredVariables["InnovationAtCurrentAnalysis"].store( _Innovation )
        # ---> avec current state
        if selfA._parameters["StoreInternalVariables"] \
            or selfA._toStore("CurrentState"):
            selfA.StoredVariables["CurrentState"].store( Xn )
        if selfA._toStore("ForecastState"):
            selfA.StoredVariables["ForecastState"].store( Xn_predicted )
        if selfA._toStore("ForecastCovariance"):
            selfA.StoredVariables["ForecastCovariance"].store( Pn_predicted )
        if selfA._toStore("BMA"):
            selfA.StoredVariables["BMA"].store( Xn_predicted - Xa )
        if selfA._toStore("InnovationAtCurrentState"):
            selfA.StoredVariables["InnovationAtCurrentState"].store( _Innovation )
        if selfA._toStore("SimulatedObservationAtCurrentState") \
            or selfA._toStore("SimulatedObservationAtCurrentOptimum"):
            selfA.StoredVariables["SimulatedObservationAtCurrentState"].store( HX_predicted )
        # ---> autres
        if selfA._parameters["StoreInternalVariables"] \
            or selfA._toStore("CostFunctionJ") \
            or selfA._toStore("CostFunctionJb") \
            or selfA._toStore("CostFunctionJo") \
            or selfA._toStore("CurrentOptimum") \
            or selfA._toStore("APosterioriCovariance"):
            Jb  = float( 0.5 * (Xa - Xb).T @ (BI @ (Xa - Xb)) )
            Jo  = float( 0.5 * _Innovation.T @ (RI @ _Innovation) )
            J   = Jb + Jo
            selfA.StoredVariables["CostFunctionJb"].store( Jb )
            selfA.StoredVariables["CostFunctionJo"].store( Jo )
            selfA.StoredVariables["CostFunctionJ" ].store( J )
            #
            if selfA._toStore("IndexOfOptimum") \
                or selfA._toStore("CurrentOptimum") \
                or selfA._toStore("CostFunctionJAtCurrentOptimum") \
                or selfA._toStore("CostFunctionJbAtCurrentOptimum") \
                or selfA._toStore("CostFunctionJoAtCurrentOptimum") \
                or selfA._toStore("SimulatedObservationAtCurrentOptimum"):
                IndexMin = numpy.argmin( selfA.StoredVariables["CostFunctionJ"][nbPreviousSteps:] ) + nbPreviousSteps
            if selfA._toStore("IndexOfOptimum"):
                selfA.StoredVariables["IndexOfOptimum"].store( IndexMin )
            if selfA._toStore("CurrentOptimum"):
                selfA.StoredVariables["CurrentOptimum"].store( selfA.StoredVariables["Analysis"][IndexMin] )
            if selfA._toStore("SimulatedObservationAtCurrentOptimum"):
                selfA.StoredVariables["SimulatedObservationAtCurrentOptimum"].store( selfA.StoredVariables["SimulatedObservationAtCurrentAnalysis"][IndexMin] )
            if selfA._toStore("CostFunctionJbAtCurrentOptimum"):
                selfA.StoredVariables["CostFunctionJbAtCurrentOptimum"].store( selfA.StoredVariables["CostFunctionJb"][IndexMin] )
            if selfA._toStore("CostFunctionJoAtCurrentOptimum"):
                selfA.StoredVariables["CostFunctionJoAtCurrentOptimum"].store( selfA.StoredVariables["CostFunctionJo"][IndexMin] )
            if selfA._toStore("CostFunctionJAtCurrentOptimum"):
                selfA.StoredVariables["CostFunctionJAtCurrentOptimum" ].store( selfA.StoredVariables["CostFunctionJ" ][IndexMin] )
        if selfA._toStore("APosterioriCovariance"):
            selfA.StoredVariables["APosterioriCovariance"].store( Pn )
        if selfA._parameters["EstimationOf"] == "Parameters" \
            and J < previousJMinimum:
            previousJMinimum    = J
            XaMin               = Xa
            if selfA._toStore("APosterioriCovariance"):
                covarianceXaMin = selfA.StoredVariables["APosterioriCovariance"][-1]
    #
    # Stockage final supplémentaire de l'optimum en estimation de paramètres
    # ----------------------------------------------------------------------
    if selfA._parameters["EstimationOf"] == "Parameters":
        selfA.StoredVariables["CurrentIterationNumber"].store( len(selfA.StoredVariables["Analysis"]) )
        selfA.StoredVariables["Analysis"].store( XaMin )
        if selfA._toStore("APosterioriCovariance"):
            selfA.StoredVariables["APosterioriCovariance"].store( covarianceXaMin )
        if selfA._toStore("BMA"):
            selfA.StoredVariables["BMA"].store( numpy.ravel(Xb) - numpy.ravel(XaMin) )
    #
    return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This file is part of SALOME ADAO module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: André Ribes, andre.ribes@edf.fr, EDF R&D

"""
This python file provides the implementation of the interface of the GUI part of
this SALOME module. This interface is required by the SALOME GUI framework.
We use here a proxy module named under the alias GuiImpl for at least three
reasons:
1. To Keep the required interface as clear as possible in this file;
2. The concrete implementation can be substituted by an alternative version;
3. We could mix several concrete implementations provided by different proxy
   modules, for example for test purposes.
"""

import adao
if not hasattr(adao, "adao_py_dir"):
    if hasattr(adao, "__file__"):
        lieu = " A potential candidate file is %s."%repr(adao.__file__)
    else:
        lieu = ""
    raise ImportError("\n\nFailed to activate module ADAO. Is it perhaps because you own a personnal perturbating file \'adao.py\' somewhere in your PATH?%s Rename, remove or move it before retrying to launch SALOME/ADAO.\n"%lieu)
from daGuiImpl import ADAOGUI_impl as GuiImpl
from daUtils import adaoLogger

adaoLogger.debug("Import ADAOGUI")

# called when module is initialized
# perform initialization actions
def initialize():
  adaoLogger.debug("initialize")
  GuiImpl.initialize()

# called when module is initialized
# return map of popup windows to be used by the module
def windows():
  adaoLogger.debug("windows")
  return GuiImpl.windows()

def views():
  adaoLogger.debug("views")
  return GuiImpl.views()

# called when module is initialized
# export module's preferences
def createPreferences():
  adaoLogger.debug("createPreferences")
  GuiImpl.createPreferences()

# called when module is activated
# returns True if activating is successfull and False otherwise
def activate():
  adaoLogger.debug("activate")
  return GuiImpl.activate()

# called when module is deactivated
def deactivate():
  adaoLogger.debug("deactivate")
  GuiImpl.deactivate()

# called when active study is changed
# active study ID is passed as parameter
#~ def activeStudyChanged( studyID ):
  #~ adaoLogger.debug("activeStudyChanged")
  #~ GuiImpl.activeStudyChanged( studyID )

# called when popup menu is invoked
# popup menu and menu context are passed as parameters
def createPopupMenu( popup, context ):
  adaoLogger.debug("createPopupMenu")
  GuiImpl.createPopupMenu(popup, context )

# called when GUI action is activated
# action ID is passed as parameter
def OnGUIEvent(commandID) :
  adaoLogger.debug("OnGUIEvent")
  GuiImpl.OnGUIEvent(commandID)

# called when module's preferences are changed
# preference's resources section and setting name are passed as parameters
def preferenceChanged( section, setting ):
  adaoLogger.debug("preferenceChanged")
  GuiImpl.preferenceChanged( section, setting )

def activeViewChanged(myView):
  adaoLogger.debug("activeViewChanged")
  GuiImpl.activeViewChanged(myView)
